import asyncio
import itertools
import json
import pathlib
from typing import Optional, List

import requests
from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta import commands
from spinta.backends import Backend
from spinta.backends.helpers import validate_and_return_transaction
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import process_stream, count_rows
from spinta.cli.helpers.errors import cli_error, ErrorCounter
from spinta.cli.helpers.export.components import CounterManager
from spinta.cli.helpers.export.helpers import validate_and_return_shallow_backend, validate_and_return_formatter, \
    export_data, filter_models_by_access_verbose
from spinta.cli.helpers.manifest import convert_str_to_manifest_path
from spinta.cli.helpers.push.utils import extract_dependant_nodes
from spinta.cli.helpers.store import prepare_manifest, attach_backends, attach_keymaps
from spinta.client import get_client_credentials, get_access_token
from spinta.commands.write import write
from spinta.components import Mode, Config, Action
from spinta.core.context import configure_context
from spinta.core.enums import Access
from spinta.datasets.keymaps.sync import sync_keymap
from spinta.exceptions import NodeNotFound
from spinta.formats.components import Format
from spinta.types.namespace import sort_models_by_ref_and_base
from spinta.utils.enums import get_enum_by_name


def import_(
    ctx: TyperContext,
    source: str,
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
    limit: Optional[int] = Option(None, help=(
        "Limit number of rows read from source."
    )),
):
    """Import data from a file"""
    context = ctx.obj
    store = prepare_manifest(context)
    manifest = store.manifest
    root = commands.get_namespace(context, manifest, '')

    with context:
        require_auth(context, auth)
        context.attach('transaction', validate_and_return_transaction, context, manifest.backend, write=True)
        with open(source) as f:
            stream = (json.loads(line.strip()) for line in f)
            stream = itertools.islice(stream, limit) if limit else stream
            stream = write(context, root, stream, changed=True)
            coro = process_stream(source, stream)
            asyncio.get_event_loop().run_until_complete(coro)


def export_(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
    backend: Optional[str] = Option(None, '-b', '--backend', help=(
        "Backend format that will import the exported data ('postgresql', 'sql', ...)"
    )),
    fmt: Optional[str] = Option(None, '-f', '--format', help=(
        "Output format ('csv', 'html, 'json', ...)"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output path"
    )),
    mode: Mode = Option('external', help=(
        "Mode of backend operation, default: external"
    )),
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Extract only specified dataset"
    )),
    no_progress_bar: bool = Option(False, '--no-progress-bar', help=(
        "Skip counting total rows to improve performance."
    )),
    credentials: str = Option(None, '--credentials', help=(
        "Credentials file, defaults to {config_path}/credentials.cfg."
    )),
    input_source: str = Option(None, '-i', '--input', help=(
        "Input source used for synchronizing keymap (credentials entry), if not given, it will skip synchronization step."
    )),
    ignore_sync_cid: bool = Option(False, '--sync-full', help=(
        "Ignores sync keymap optimization and fully synchronizes data (does not start from last synchronized row)."
    )),
    max_error_count: int = Option(50, '--max-errors', help=(
        "If errors exceed given number, export command will be stopped."
    )),
    read_timeout: float = Option(300, '--read-timeout', help=(
        "Timeout for reading a response, default: 5 minutes (300s). The value is in seconds."
    )),
    connect_timeout: float = Option(5, '--connect-timeout', help=(
        "Timeout for connecting, default: 5 seconds."
    )),
    access: str = Option('private', help=(
        "Export properties with at least specified access"
    )),
):
    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context, full_load=True)
    config: Config = context.get('config')

    if backend and fmt:
        cli_error(
            "Export can only output to one type (either `--backend` or `--format`), but not both."
        )

    if not backend and not fmt:
        cli_error(
            "Export must be given an output format (either `--backend` or `--format`)."
        )

    if backend:
        backend: Backend = validate_and_return_shallow_backend(context, backend)

    if fmt:
        fmt: Format = validate_and_return_formatter(context, fmt)

    manifest = store.manifest
    if dataset and not commands.has_dataset(context, manifest, dataset):
        cli_error(
            str(NodeNotFound(manifest, type='dataset', name=dataset))
        )

    if not output:
        cli_error(
            "Output argument is required (`--output`)."
        )
    commands.validate_export_output(context, fmt or backend, output)
    access = get_enum_by_name(Access, access)

    with context:
        require_auth(context)
        error_counter = ErrorCounter(max_count=max_error_count)

        attach_backends(context, store, manifest)
        attach_keymaps(context, store)
        ns = commands.get_namespace(context, manifest, '')
        models = commands.traverse_ns_models(context, ns, manifest, Action.SEARCH, dataset_=dataset, source_check=True)
        models = filter_models_by_access_verbose(models, access)
        models = sort_models_by_ref_and_base(list(models))

        # Synchronize only if input_source is given
        if input_source:
            if credentials:
                credsfile = pathlib.Path(credentials)
                if not credsfile.exists():
                    cli_error(
                        f"Credentials file {credsfile} does not exit."
                    )
            else:
                credsfile = config.credentials_file
            creds = get_client_credentials(credsfile, input_source)
            echo(f"Get access token from {creds.server}")
            token = get_access_token(creds)

            client = requests.Session()
            client.headers['Content-Type'] = 'application/json'
            client.headers['Authorization'] = f'Bearer {token}'

            # Synchronize keymaps
            with manifest.keymap as km:
                dependant_models = extract_dependant_nodes(context, models, True)
                sync_keymap(
                    context=context,
                    keymap=km,
                    client=client,
                    server=creds.server,
                    models=dependant_models,
                    error_counter=error_counter,
                    no_progress_bar=no_progress_bar,
                    reset_cid=ignore_sync_cid,
                    timeout=(connect_timeout, read_timeout)
                )
        else:
            dependant_models = extract_dependant_nodes(context, models, True)
            if dependant_models:
                cli_error((
                    f"Detected some models, that might require synchronization step: {', '.join([model.model_type() for model in dependant_models])}\n"
                    "Add --input argument to run export with incremental synchronization."
                ))
            echo("Input source not given, skipping synchronization step.")

        counts = count_rows(
            context,
            models,
            no_progress_bar=no_progress_bar
        )

        with CounterManager(
            enabled=not no_progress_bar,
            totals=counts
        ) as counter:
            asyncio.get_event_loop().run_until_complete(
                export_data(
                    context,
                    models,
                    fmt or backend,
                    output,
                    counter,
                    access
                )
            )



from collections import abc
from typing import Any, Union

from multipledispatch import dispatch

from spinta.backends import Backend
from spinta.components import Context, Property
from spinta.core.ufuncs import Expr
from spinta.dimensions.enum.helpers import get_prop_enum
from spinta.exceptions import ValueNotInEnum
from spinta.ufuncs.querybuilder.components import Selected
from spinta.ufuncs.resultbuilder.components import ResultBuilder
from spinta.utils.schema import NA

ResultBuilderGetter = abc.Callable[[], ResultBuilder]


def _resolve_expr(context: Context, row: Any, sel: Selected, result_builder_getter: Union[ResultBuilderGetter, ResultBuilder]) -> Any:
    if sel.item is None:
        val = None
    else:
        val = row[sel.item]

    result_builder = result_builder_getter
    if isinstance(result_builder, abc.Callable):
        result_builder = result_builder()
    env = result_builder.init(val, sel.prop, row)
    return env.resolve(sel.prep)


def _aggregate_values(data: Any, target: Property):
    key_path = target.place
    key_parts = key_path.split('.')

    # Drop first part, since if nested prop is part of the list will always be first value
    # ex: from DB we get {"notes": [{"note": 0}]}
    # but after fetching the value we only get [{"note": 0}]
    # so if our place is "notes.note", we need to drop "notes" part
    if len(key_parts) > 1:
        key_parts = key_parts[1:]

    def recursive_collect(sub_data, depth=0):
        if depth < len(key_parts):
            if isinstance(sub_data, list):
                collected = []
                for item in sub_data:
                    collected.extend(recursive_collect(item, depth))
                return collected
            elif isinstance(sub_data, dict) and key_parts[depth] in sub_data:
                return recursive_collect(sub_data[key_parts[depth]], depth + 1)
        else:
            return [sub_data]

        return []

    # Start the recursive collection process
    return recursive_collect(data, 0)


# This is the fastest way to call this function, use this if you do not want to recreate `ResultBuilder`
@dispatch(Context, ResultBuilder, object, object)
def get_row_value(context: Context, result_builder: ResultBuilder, row: Any, sel: Any) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        True,
        result_builder
    )


# This is the fastest way to call this function, use this if you do not want to recreate `ResultBuilder`
@dispatch(Context, ResultBuilder, object, object, bool)
def get_row_value(context: Context, result_builder: ResultBuilder, row: Any, sel: Any, check_enums: bool) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        check_enums,
        result_builder
    )


# This is the second-fastest way to call this function, it allows for `ResultBuilder` recreation
@dispatch(Context, abc.Callable, object, object, bool)
def get_row_value(context: Context, result_builder_getter: ResultBuilderGetter, row: Any, sel: Any, check_enums: bool) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        check_enums,
        result_builder_getter
    )


# This is the second-fastest way to call this function, it allows for `ResultBuilder` recreation
@dispatch(Context, abc.Callable, object, object)
def get_row_value(context: Context, result_builder_getter: ResultBuilderGetter, row: Any, sel: Any) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        True,
        result_builder_getter
    )


# This is the slowest way to call this function, use this only if you cannot pass `backend_result_builder_getter` directly
@dispatch(Context, Backend, object, object)
def get_row_value(context: Context, backend: Backend, row: Any, sel: Any) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        True,
        backend_result_builder_getter(context, backend)
    )


# This is the slowest way to call this function, use this only if you cannot pass `backend_result_builder_getter` directly
@dispatch(Context, Backend, object, object, bool)
def get_row_value(context: Context, backend: Backend, row: Any, sel: Any, check_enums: bool) -> Any:
    return _get_row_value(
        context,
        row,
        sel,
        check_enums,
        backend_result_builder_getter(context, backend)
    )


def backend_result_builder_getter(context: Context, backend: Backend) -> ResultBuilderGetter:
    return lambda: backend.result_builder_class(context)


def _get_row_value(
    context: Context,
    row: Any,
    sel: Any,
    check_enums: bool,
    result_builder_getter: Union[ResultBuilderGetter, ResultBuilder]
):
    if isinstance(sel, Selected):
        if isinstance(sel.prep, Expr):
            val = _resolve_expr(context, row, sel, result_builder_getter)
        elif sel.prep is not NA:
            val = _get_row_value(context, row, sel.prep, check_enums, result_builder_getter)
        else:
            if sel.item is not None:
                val = row[sel.item]
                prop = sel.prop
                if prop is not None and prop.list is not None:
                    val = _aggregate_values(val, prop)
            else:
                val = None

        if check_enums:
            if enum := get_prop_enum(sel.prop):
                if val is None:
                    pass
                elif str(val) in enum:
                    item = enum[str(val)]
                    if item.prepare is not NA:
                        val = item.prepare
                else:
                    raise ValueNotInEnum(sel.prop, value=val)

        return val
    if isinstance(sel, tuple):
        return tuple(_get_row_value(context, row, v, check_enums, result_builder_getter) for v in sel)
    if isinstance(sel, list):
        return [_get_row_value(context, row, v, check_enums, result_builder_getter) for v in sel]
    if isinstance(sel, dict):
        return {k: _get_row_value(context, row, v, check_enums, result_builder_getter) for k, v in sel.items()}
    return sel



import csv
import pathlib
from typing import Any
from typing import Dict
from typing import Iterator

from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import StreamingResponse

from spinta import commands
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.formats.csv.components import Csv
from spinta.formats.csv.components import IterableFile
from spinta.formats.helpers import get_model_tabular_header, rename_page_col
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.components import ManifestRow
from spinta.manifests.tabular.constants import DATASET
from spinta.manifests.tabular.helpers import datasets_to_tabular, write_csv
from spinta.utils.nestedstruct import flatten, sepgetter
from spinta.utils.response import aiter


@commands.render.register(Context, Request, Model, Csv)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Csv,
    *,
    action: Action,
    params: UrlParams,
    data: Iterator[Dict[str, Any]],
    status_code: int = 200,
    headers: Dict[str, str] = None,
) -> Response:
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{model.basename}.csv"'
    return StreamingResponse(
        aiter(_render_model_csv(context, model, action, params, data)),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


@commands.render.register(Context, Manifest, Csv)
def render(
    context: Context,
    manifest: Manifest,
    fmt: Csv,
    *,
    action: Action = None,
    params: UrlParams = None,
    status_code: int = 200,
    headers: Dict[str, str] = None,
    path: str = None
) -> Response:
    rows = datasets_to_tabular(context, manifest)
    rows = ({c: row[c] for c in DATASET} for row in rows)
    if not path:
        headers = headers or {}
        headers['Content-Disposition'] = f'attachment; filename="{manifest.name}.csv"'
        return StreamingResponse(
            aiter(_render_manifest_csv(rows)),
            status_code=status_code,
            media_type=fmt.content_type,
            headers=headers,
        )
    else:
        write_csv(pathlib.Path(path), rows, DATASET)


def _render_manifest_csv(
    rows: Iterator[ManifestRow]
):
    stream = IterableFile()
    writer = csv.DictWriter(stream, fieldnames=DATASET)
    writer.writeheader()
    writer.writerows(rows)
    yield from stream


def _render_model_csv(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    data: Iterator[Dict[str, Any]],
):
    rows = flatten(data, sepgetter(model))
    # Rename _page to _page.next
    rows = rename_page_col(rows)
    cols = get_model_tabular_header(context, model, action, params)
    cols = [col if col != '_page' else '_page.next' for col in cols]
    stream = IterableFile()
    writer = csv.DictWriter(stream, fieldnames=cols)
    writer.writeheader()
    memory = next(rows, None)
    for row in rows:
        writer.writerow({k: v for k, v in memory.items() if k != '_page.next'})
        yield from stream
        memory = row

    if memory is not None:
        writer.writerow(memory)
    yield from stream



from spinta.core.ufuncs import ufunc, Expr
from spinta.types.geometry.components import Geometry
from spinta.ufuncs.querybuilder.components import Selected
from spinta.ufuncs.resultbuilder.components import ResultBuilder

import shapely
from shapely.geometry.base import BaseGeometry


@ufunc.resolver(ResultBuilder)
def count(env: ResultBuilder):
    pass


@ufunc.resolver(ResultBuilder, Selected, Selected)
def point(env: ResultBuilder, x: Selected, y: Selected) -> str:
    x = env.data[x.item]
    y = env.data[y.item]
    return f'POINT ({x} {y})'


@ufunc.resolver(ResultBuilder, Expr)
def split(env: ResultBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return env.call('split', *args, **kwargs)


@ufunc.resolver(ResultBuilder, str)
def split(env: ResultBuilder, separator: str):
    return env.this.split(separator)


@ufunc.resolver(ResultBuilder, Expr)
def flip(env: ResultBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return env.call('flip', *args, **kwargs)


@ufunc.resolver(ResultBuilder)
def flip(env: ResultBuilder):
    return env.call('flip', env.prop.dtype, env.this)


@ufunc.resolver(ResultBuilder, Geometry, str)
def flip(env: ResultBuilder, dtype: Geometry, value: str):
    values = value.split(';', 1)
    shape: BaseGeometry = shapely.wkt.loads(values[-1])
    inverted: BaseGeometry = shapely.ops.transform(lambda x, y: (y, x), shape)
    inverted: str = shapely.wkt.dumps(inverted, trim=True)
    if len(values) > 1:
        return ';'.join([*values[:-1], inverted])
    return inverted


@ufunc.resolver(ResultBuilder, Expr)
def swap(env: ResultBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return env.call('swap', *args, *kwargs)



from spinta.components import Property, PageInfo
from spinta.core.ufuncs import ufunc, Expr, Pair, Bind, Positive, Negative
from spinta.exceptions import InvalidArgumentInExpression, FieldNotInResource
from spinta.ufuncs.loadbuilder.components import LoadBuilder


@ufunc.resolver(LoadBuilder, Expr, name='page')
def page_(env, expr):
    args = env.resolve(expr.args)
    page = PageInfo(env.model)
    if len(args) > 0:
        for item in args:
            res = env.resolve(item)
            res = env.call('page_item', res)
            if isinstance(res[1], Property):
                page.keys[res[0]] = res[1]
            else:
                if res[0] == 'size' and isinstance(res[1], int):
                    page.size = res[1]
                else:
                    raise InvalidArgumentInExpression(arguments=res[0], expr='page')
    else:
        page.enabled = False
    return page


@ufunc.resolver(LoadBuilder, Pair, name='page_item')
def page_item(env, field):
    return field.name, field.value


@ufunc.resolver(LoadBuilder, Bind, name='page_item')
def page_item(env, field):
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return prop.name, prop


@ufunc.resolver(LoadBuilder, Positive, name='page_item')
def page_item(env, field):
    return env.call('page_item', Bind(field.name))


@ufunc.resolver(LoadBuilder, Negative, name='page_item')
def page_item(env, field):
    resolved = env.call('page_item', Bind(field.name))
    return f'-{resolved[0]}', resolved[1]


@ufunc.resolver(LoadBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return args


@ufunc.resolver(LoadBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return args



from typing import Optional
import sqlalchemy as sa

from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.internal_sql.components import InternalSQLManifest
from spinta.manifests.internal_sql.helpers import get_table_structure


@commands.configure.register(Context, InternalSQLManifest)
def configure(context: Context, manifest: InternalSQLManifest):
    rc: RawConfig = context.get('rc')
    path: Optional[str] = rc.get('manifests', manifest.name, 'path')
    manifest.path = path
    url = sa.engine.make_url(manifest.path)
    engine = sa.create_engine(url)
    manifest.engine = engine
    meta = sa.MetaData(engine)
    manifest.table = get_table_structure(meta)



from typing import Any

from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.backends.mssql.helpers import group_array
from spinta.datasets.backends.sql.backends.mssql.ufuncs.query.components import MSSQLQueryBuilder


@ufunc.resolver(MSSQLQueryBuilder, object)
def _group_array(env: MSSQLQueryBuilder, columns: Any):
    return group_array(columns)



import contextlib

from spinta.datasets.backends.dataframe.components import DaskBackend


class Json(DaskBackend):
    type: str = 'dask/json'

    @contextlib.contextmanager
    def begin(self):
        yield



import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import overload, Iterator, List, Tuple, Callable

from starlette.requests import Request
from starlette.responses import FileResponse
from starlette.responses import Response

from spinta import commands
from spinta.accesslog import AccessLog
from spinta.accesslog import log_response
from spinta.backends.components import Backend
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.backends.nobackend.components import NoBackend
from spinta.compat import urlparams_to_expr
from spinta.components import Context, Node, Action, UrlParams, Page, get_page_size, pagination_enabled
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.exceptions import ItemDoesNotExist
from spinta.exceptions import UnavailableSubresource, InfiniteLoopWithPagination, BackendNotGiven, TooShortPageSize, \
    TooShortPageSizeKeyRepetition
from spinta.renderer import render
from spinta.types.datatype import DataType
from spinta.types.datatype import File
from spinta.types.datatype import Object
from spinta.ufuncs.querybuilder.components import QueryParams
from spinta.ufuncs.querybuilder.helpers import update_query_with_url_params, add_page_expr
from spinta.utils.data import take


@commands.getall.register(Context, Model, Request)
async def getall(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)
    backend = model.backend
    if isinstance(backend, NoBackend):
        raise BackendNotGiven(model)

    copy_page = commands.create_page(model.page, params)

    is_page_enabled = copy_page and copy_page and copy_page.enabled
    expr = urlparams_to_expr(params)
    query_params = QueryParams()
    update_query_with_url_params(query_params, params)
    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
    )

    func_properties = None
    if params.select_funcs:
        func_properties = {key: func.prop for key, func in params.select_funcs.items()}

    if params.head:
        rows = []
    else:
        if is_page_enabled:
            rows = commands.getall(
                context,
                model,
                copy_page,
                query=expr,
                limit=params.limit,
                default_expand=False,
                params=query_params,
                extra_properties=func_properties
            )
        else:
            rows = commands.getall(
                context,
                model,
                backend,
                params=query_params,
                query=expr,
                default_expand=False,
                extra_properties=func_properties
            )

    rows = prepare_data_for_response(
        context,
        model,
        action,
        params,
        rows,
    )

    rows = log_response(context, rows)

    return render(context, request, model, params, rows, action=action)


@commands.getall.register(Context, Model, Page)
def getall(
    context: Context,
    model: Model,
    page: Page,
    *,
    query: Expr = None,
    limit: int = None,
    **kwargs
) -> Iterator:
    backend = model.backend
    if isinstance(backend, NoBackend):
        raise BackendNotGiven(model)

    config = context.get('config')
    size = get_page_size(config, model, page)

    # Add 1 to see future value (to see if it finished, check for infinite loops and page size miss matches).
    page.size = size + 1

    page_meta = PaginationMetaData(
        page_size=size,
        limit=limit
    )
    while not page_meta.is_finished:
        page_meta.is_finished = True
        query = add_page_expr(query, page)
        rows = commands.getall(
            context,
            model,
            backend,
            query=query,
            **kwargs
        )

        yield from get_paginated_values(page, page_meta, rows, extract_source_page_keys)


def prepare_data_for_response(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    rows,
    reserved: List[str] = None,
):
    if isinstance(rows, dict):
        rows = [rows]

    prop_select = params.select_props
    func_select = params.select_funcs

    prop_select_tree = get_select_tree(context, action, prop_select)
    func_select_tree = get_select_tree(context, action, func_select)

    if reserved is None:
        if action == Action.SEARCH:
            reserved = ['_type', '_id', '_revision', '_base']
        else:
            reserved = ['_type', '_id', '_revision']
        if pagination_enabled(model, params):
            reserved.append('_page')
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        prop_select_tree,
        reserved=reserved,
        include_denorm_props=False,
    )

    for row in rows:
        result = commands.prepare_data_for_response(
            context,
            model,
            params.fmt,
            row,
            action=action,
            select=prop_select_tree,
            prop_names=prop_names,
        )
        if func_select:
            for key, func_prop in func_select.items():
                result[key] = commands.prepare_dtype_for_response(
                    context,
                    params.fmt,
                    func_prop.prop.dtype,
                    row[key],
                    data=row,
                    action=action,
                    select=func_select_tree
                )
        yield result


def _is_iter_last_real_value(it: int, total: int, added_size: int = 1):
    return it > (total - 1 - added_size)


def _is_iter_last_potential_value(it: int, total: int):
    return it > (total - 1)


def extract_source_page_keys(row: dict):
    if '_page' in row:
        return row['_page']
    return []


@dataclass
class PaginationMetaData:
    page_size: int
    previous_first_value = None
    is_first_iter = True
    is_finished = False
    true_count = 0
    limit: int = None

    def handle_count(self) -> bool:
        if self.limit:
            if self.true_count >= self.limit:
                self.is_finished = True
                return True
            self.true_count += 1
        return False


def get_paginated_values(model_page: Page, meta: PaginationMetaData, rows, extract_page_keys: Callable):
    size = meta.page_size

    if model_page.first_time != meta.is_first_iter:
        model_page.first_time = meta.is_first_iter
    if meta.is_first_iter:
        meta.is_first_iter = False

    initial_key = [val.value for val in model_page.by.values()]
    current_first_value = None
    previous_value = None
    key_repetition = [[], 0]
    flag_for_potential_key_repetition = False
    for i, row in enumerate(rows):
        keys = extract_page_keys(row).copy()
        # Check if future value is not the same as last value
        if key_repetition[0] == keys:
            key_repetition[1] += 1
        else:
            key_repetition = [keys, 0]

        if _is_iter_last_potential_value(i, size):
            meta.is_finished = False
            if flag_for_potential_key_repetition:
                raise TooShortPageSize(
                    model_page,
                    page_size=size,
                    page_values=previous_value
                )
            if key_repetition[1] > 0:
                raise TooShortPageSizeKeyRepetition(
                    model_page,
                    page_size=size,
                    page_values=previous_value,
                )
            break

        previous_value = row

        limit_result = meta.handle_count()
        if limit_result:
            break

        # Check if row is completely the same as previous page first row
        if current_first_value is None:
            current_first_value = row
            if current_first_value == meta.previous_first_value:
                raise InfiniteLoopWithPagination(
                    model_page,
                    page_size=size,
                    page_values=current_first_value
                )
            meta.previous_first_value = current_first_value

        model_page.update_values_from_list(keys)

        # Check if initial key is the same as last key
        if _is_iter_last_real_value(i, size):
            last_key = keys
            if initial_key == last_key:
                flag_for_potential_key_repetition = True

        yield row


def _update_expr_args(
    expr: Expr,
    name: str,
    args: List,
    override: bool = False
) -> Tuple[bool, Expr]:
    updated = False
    expr.args = list(expr.args)
    if expr.name == name:
        if override:
            expr.args = args
            updated = True
        else:
            if expr.args:
                expr.args.extend(args)
                updated = True
    else:
        for i, arg in enumerate(expr.args):
            if isinstance(arg, Expr):
                updated, expr.args[i] = _update_expr_args(arg, name, args, override)
                if updated:
                    break
    expr.args = tuple(expr.args)
    return updated, expr


@overload
@commands.getone.register(Context, Request, Node)
async def getone(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    if params.prop and params.propref:
        resp = await commands.getone(
            context,
            request,
            params.prop,
            params.prop.dtype,
            params.model.backend,
            action=action,
            params=params,
        )
    elif params.prop:
        resp = await commands.getone(
            context,
            request,
            params.prop,
            params.prop.dtype,
            params.prop.dtype.backend or params.model.backend,
            action=action,
            params=params,
        )
    else:
        resp = await commands.getone(
            context,
            request,
            params.model,
            params.model.backend,
            action=action,
            params=params,
        )

    accesslog: AccessLog = context.get('accesslog')
    accesslog.response(objects=1)

    return resp


@overload
@commands.getone.register(Context, Request, Property, DataType, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    raise UnavailableSubresource(prop=prop.name, prop_type=prop.dtype.name)


@overload
@commands.getone.register(Context, Request, Model, Backend)
async def getone(
    context: Context,
    request: Request,
    model: Model,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
        id_=params.pk,
    )

    data = commands.getone(context, model, backend, id_=params.pk)
    data = next(prepare_data_for_response(context, model, action, params, data, reserved=[]))
    return render(context, request, model, params, data, action=action)


@overload
@commands.getone.register(Context, Request, Property, Object, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: Object,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    data = commands.getone(
        context,
        prop,
        dtype,
        backend,
        id_=params.pk,
    )

    data = commands.prepare_data_for_response(
        context,
        prop.dtype,
        params.fmt,
        data,
        action=action,
    )
    return render(context, request, prop, params, data, action=action)


@overload
@commands.getone.register(Context, Request, Property, File, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    # Get metadata from model backend
    data = commands.getone(
        context,
        prop,
        dtype,
        prop.model.backend,
        id_=params.pk,
    )

    # Return file metadata from model backend
    if params.propref:
        data = commands.prepare_data_for_response(
            context,
            prop.dtype,
            params.fmt,
            data,
            action=Action.GETONE,
        )
        return render(context, request, prop, params, data, action=action)

    # Return file content from property backend
    else:
        value = take(prop.place, data)

        file = commands.getfile(
            context,
            prop,
            dtype,
            dtype.backend,
            data=value,
        )

        if file is None:
            raise ItemDoesNotExist(dtype, id=params.pk)

        filename = value['_id']

        if isinstance(file, bytes):
            ResponseClass = Response
        elif isinstance(file, Path):
            ResponseClass = FileResponse
        else:
            raise ValueError(f"Unknown file type {type(file)}.")

        return ResponseClass(
            file,
            media_type=value.get('_content_type'),
            headers={
                'Revision': data['_revision'],
                'Content-Disposition': (
                    f'attachment; filename="{filename}"'
                    if filename else
                    'attachment'
                )
            },
        )


@commands.changes.register(Context, Model, Request)
async def changes(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)

    if params.head:
        rows = []
    else:
        accesslog = context.get('accesslog')
        accesslog.request(
            # XXX: Read operations does not have a transaction, but it
            #      is needed for loging.
            txn=str(uuid.uuid4()),
            model=model.model_type(),
            action=action.value,
        )

        rows = commands.changes(
            context,
            model,
            model.backend,
            id_=params.pk,
            limit=params.limit,
            offset=params.changes_offset,
        )

    rows = prepare_data_for_response(context, model, action, params, rows, reserved=[
        '_cid',
        '_created',
        '_op',
        '_id',
        '_txn',
        '_revision',
    ])

    return render(context, request, model, params, rows, action=action)


@commands.summary.register(Context, Request, Model)
async def summary(
    context: Context,
    request: Request,
    model: Model,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)
    backend = model.backend

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
    )
    if params.head:
        rows = []
    else:
        expr = urlparams_to_expr(params)
        rows = commands.summary(context, model, backend, expr)
    rows = log_response(context, rows)
    return render(context, request, model, params, rows, action=action)



from spinta import commands, exceptions
from spinta.backends import Backend
from spinta.components import Context, Config
from spinta.core.config import RawConfig
from spinta.datasets.backends.dataframe.backends.json.components import Json
from spinta.datasets.backends.dataframe.backends.xml.components import Xml
from spinta.datasets.backends.sql.components import Sql
from spinta.manifests.dict.components import JsonManifest, XmlManifest
from spinta.manifests.sql.components import SqlManifest
from spinta.manifests.tabular.components import CsvManifest
from spinta.datasets.backends.dataframe.backends.csv.components import Csv
from spinta.manifests.helpers import get_manifest_from_type


@commands.backend_to_manifest_type.register(Context, str)
def backend_to_manifest_type(context: Context, backend_type: str):
    config: Config = context.get('config')

    if backend_type not in config.components['backends']:
        raise exceptions.BackendNotFound(name=backend_type)
    Backend_ = config.components['backends'][backend_type]
    backend: Backend = Backend_()
    backend.type = backend_type
    return commands.backend_to_manifest_type(context, backend)


@commands.backend_to_manifest_type.register(Context, Backend)
def backend_to_manifest_type(context: Context, backend: Backend):
    rc: RawConfig = context.get('rc')
    return get_manifest_from_type(rc, backend.type)


@commands.backend_to_manifest_type.register(Context, Csv)
def backend_to_manifest_type(context: Context, backend: Csv):
    return CsvManifest


@commands.backend_to_manifest_type.register(Context, Json)
def backend_to_manifest_type(context: Context, backend: Json):
    return JsonManifest


@commands.backend_to_manifest_type.register(Context, Xml)
def backend_to_manifest_type(context: Context, backend: Xml):
    return XmlManifest


@commands.backend_to_manifest_type.register(Context, Sql)
def backend_to_manifest_type(context: Context, backend: Sql):
    return SqlManifest



from spinta import commands
from spinta.components import Context, Property
from spinta.exceptions import InvalidDenormProperty
from spinta.types.datatype import Denorm, Ref, ExternalRef


@commands.check.register(Context, Denorm)
def check(context: Context, dtype: Denorm):
    parent = dtype.prop.parent
    if isinstance(parent, Property) and isinstance(parent.dtype, ExternalRef):
        if dtype.rel_prop in parent.dtype.refprops:
            raise InvalidDenormProperty(dtype, denorm=dtype.prop.name, ref=parent.name, refprops=[prop.name for prop in parent.dtype.refprops])




from typing import List

from spinta.types.text.components import Text


def determine_language_property_for_text(dtype: Text, prioritized_langs: list, default_langs: list):
    existing_langs = list(dtype.langs.keys())
    selected_lang = determine_langauge_for_text(existing_langs, prioritized_langs, default_langs)
    return dtype.langs[selected_lang]


def determine_langauge_for_text(existing_langs: List[str], prioritized_langs: List[str], default_langs: List[str]):
    selected_lang = None
    if prioritized_langs:
        for lang in prioritized_langs:
            if lang in existing_langs:
                selected_lang = lang
                break
    if not selected_lang and default_langs:
        for lang in default_langs:
            if lang in existing_langs:
                selected_lang = lang
                break
    if not selected_lang:
        selected_lang = existing_langs[0]
    return selected_lang



import json
from typing import Any

from spinta import commands, spyna
from spinta.components import Property, Model, Context, Mode
from spinta.core.ufuncs import Env, asttoexpr
from spinta.datasets.keymaps.components import KeyMap
from spinta.exceptions import GivenValueCountMissmatch, MultiplePrimaryKeyCandidatesFound, NoPrimaryKeyCandidatesFound
from spinta.types.datatype import Ref, Array
from spinta.types.namespace import check_if_model_has_backend_and_source
from spinta.utils.schema import NA


def handle_ref_key_assignment(context: Context, keymap: KeyMap, env: Env, value: Any, ref: Ref) -> dict:
    keymap_name = ref.model.model_type()
    if ref.refprops != ref.model.external.pkeys:
        keymap_name = f'{keymap_name}.{"_".join(prop.name for prop in ref.refprops)}'

    if not isinstance(value, (tuple, list)):
        value = [value]
    else:
        value = list(value)

    prop_count_mapping = {}
    for prop in ref.refprops:
        if isinstance(prop.dtype, Array) and env:
            items = env.resolve(prop.external.prepare)
            prop_count_mapping[prop.name] = len(items) if items is not NA else 1
        else:
            prop_count_mapping[prop.name] = 1
    expected_count = sum(item for item in prop_count_mapping.values())
    if len(value) != expected_count:
        raise GivenValueCountMissmatch(ref, given_count=len(value), expected_count=expected_count)

    if commands.identifiable(ref.prop):
        target_value = value
        if len(value) == 1:
            target_value = value[0]

        val = None
        contains = keymap.contains(keymap_name, target_value)
        if not contains:
            if target_value is None:
                return {'_id': None}

            ref_model = ref.model

            # FIXME Quick hack when trying to get `Internal` model keys while running in `External` mode (should probably return error, or None)
            if ref_model.mode == Mode.external and not check_if_model_has_backend_and_source(ref_model):
                return {
                    '_id': keymap.encode(keymap_name, target_value)
                }

            expr_parts = ['select()']
            for i, prop in enumerate(ref.refprops):
                expr_parts.append(f'{prop.place}="{value[i]}"')
            expr = asttoexpr(spyna.parse('&'.join(expr_parts)))
            rows = commands.getall(
                context,
                ref_model,
                ref_model.backend,
                query=expr
            )

            found_value = False
            for row in rows:
                if val is not None:
                    raise MultiplePrimaryKeyCandidatesFound(ref, values=target_value)
                val = row['_id']
                found_value = True

            if not found_value:
                raise NoPrimaryKeyCandidatesFound(ref, values=target_value)
        else:
            val = keymap.encode(keymap_name, target_value)
        val = {'_id': val}
    else:
        val = {}
        i = 0
        for prop, count in prop_count_mapping.items():
            values = value[i:i + count]
            if len(values) == 1:
                values = values[0]
            val[prop] = values
            i = i + count
    return val


def generate_ref_id_using_select(context: Context, dtype: Ref, data: dict) -> str:
    ref_model = dtype.model
    expr_parts = ['select()']
    for prop in dtype.refprops:
        expr_parts.append(f'{prop.place}="{data[prop.name]}"')
    expr = asttoexpr(spyna.parse('&'.join(expr_parts)))
    rows = commands.getall(
        context,
        ref_model,
        ref_model.backend,
        query=expr
    )

    found_value = False
    val = None
    for row in rows:
        if val is not None:
            raise MultiplePrimaryKeyCandidatesFound(dtype, values=data)
        val = row['_id']
        found_value = True

    if not found_value:
        raise NoPrimaryKeyCandidatesFound(dtype, values=data)

    return val


def generate_pk_for_row(context: Context, model: Model, row: Any, keymap, pk_val: Any):
    pk = None
    if model.base and commands.identifiable(model.base):
        pk_val_base = extract_values_from_row(row, model.base.parent, model.base.pk or model.base.parent.external.pkeys)
        key = model.base.parent.model_type()
        if model.base.pk and model.base.pk != model.base.parent.external.pkeys:
            joined = '_'.join(pk.name for pk in model.base.pk)
            key = f'{key}.{joined}'
        pk = keymap.encode(key, pk_val_base)

    pk = keymap.encode(model.model_type(), pk_val, pk)
    if pk and model.required_keymap_properties:
        for combination in model.required_keymap_properties:
            joined = '_'.join(combination)
            key = f'{model.model_type()}.{joined}'
            val = extract_values_from_row(row, model, combination)
            keymap.encode(key, val, pk)
    return pk


def extract_values_from_row(row: Any, model: Model, keys: list):
    return_list = []
    for key in keys:
        if isinstance(key, str):
            key = model.properties[key]
        if isinstance(key, Property):
            if key.external and key.external.name:
                key = key.external.name
            else:
                key = key.name

        return_list.append(row[key])
    if len(return_list) == 1:
        return_list = return_list[0]
    return return_list


def handle_external_array_type(
    context: Context,
    dtype: Array,
    keymap: KeyMap,
    env: Env,
    val: Any
):
    if isinstance(val, str):
        val = json.loads(val)

    if not isinstance(val, list):
        return val

    if isinstance(dtype.items.dtype, Ref):
        return [handle_ref_key_assignment(context, keymap, env, value, dtype.items.dtype) for value in val]

    return val


def flatten_keymap_encoding_values(data: object):
    # Backwards compatibility function, that converts nested dict values to flat list values
    # Used for generating and looking up keymap values
    if isinstance(data, dict):
        if len(data) == 1:
            return flatten_keymap_encoding_values(list(data.values())[0])
        return flatten_keymap_encoding_values(list(data.values()))
    elif isinstance(data, (list, tuple)):
        return list(flatten_keymap_encoding_values(item) for item in data)
    return data



from __future__ import annotations

from spinta import commands
from spinta.backends import Backend
from spinta.components import Context, Model, Mode
from spinta.datasets.backends.helpers import generate_ref_id_using_select, flatten_keymap_encoding_values
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.keymaps.components import KeyMap
from spinta.exceptions import GivenValueCountMissmatch, KeymapValueNotFound
from spinta.types.datatype import Ref, ExternalRef
from spinta.types.namespace import check_if_model_has_backend_and_source


@commands.cast_backend_to_python.register(Context, Model, Sql, dict)
def cast_backend_to_python(
    context: Context,
    model: Model,
    backend: Sql,
    data: dict,
    *,
    keymap: KeyMap = None,
    **kwargs
):
    if keymap is None:
        keymap = context.get(f'keymap.{model.keymap.name}')

    super_ = commands.cast_backend_to_python[Context, Model, Backend, dict]
    return super_(context, model, backend, data, keymap=keymap, **kwargs)


@commands.cast_backend_to_python.register(Context, Ref, Sql, dict)
def cast_backend_to_python(
    context: Context,
    dtype: Ref,
    backend: Sql,
    data: dict,
    *,
    keymap: KeyMap = None,
    **kwargs
):
    processed_data = {}
    if keymap is None:
        keymap = context.get(f'keymap.{dtype.prop.model.keymap.name}')

    for key, value in data.items():
        # Skip '_id', it will be processed below, since it needs to be converted to uuid type
        if key == '_id':
            continue

        prop = commands.resolve_property(dtype.prop.model, f'{dtype.prop.place}.{key}')
        if prop is not None:
            processed_data[key] = commands.cast_backend_to_python(
                context,
                prop,
                backend,
                value,
                keymap=keymap,
                **kwargs
            )

    ref_model = dtype.model
    keymap_name = ref_model.model_type()
    if dtype.refprops != ref_model.external.pkeys:
        keymap_name = f'{keymap_name}.{"_".join(prop.name for prop in dtype.refprops)}'

    values = {}
    id_data = data.get('_id')
    if not id_data:
        return processed_data

    for prop in dtype.refprops:
        if prop.name not in id_data:
            # Skip trying to create _id, since not all refprop are given
            return processed_data
        values[prop.name] = (id_data[prop.name])

    encoding_values = list(values.values())
    if len(encoding_values) == 1:
        encoding_values = encoding_values[0]

    # Backwards compatibility, all nested values are converted to list values without keys
    encoding_values = flatten_keymap_encoding_values(encoding_values)
    contains = keymap.contains(keymap_name, encoding_values)

    if contains:
        id_value = keymap.encode(keymap_name, encoding_values)
    elif encoding_values is None:
        id_value = None
    elif ref_model.mode == Mode.external and not check_if_model_has_backend_and_source(ref_model):
        raise KeymapValueNotFound(
            dtype,
            keymap=keymap.name,
            model_name=dtype.model.name,
            values=encoding_values,
        )
    else:
        id_value = generate_ref_id_using_select(context, dtype, values)

    processed_data['_id'] = commands.cast_backend_to_python(
        context,
        ref_model.properties['_id'],
        backend,
        id_value,
        keymap=keymap,
        **kwargs
    )

    if not processed_data or all(value is None for value in processed_data.values()):
        return None

    return processed_data


@commands.cast_backend_to_python.register(Context, ExternalRef, Sql, dict)
def cast_backend_to_python(
    context: Context,
    dtype: ExternalRef,
    backend: Sql,
    data: dict,
    **kwargs
):
    processed_data = {}
    for key, value in data.items():
        prop = commands.resolve_property(dtype.prop.model, f'{dtype.prop.place}.{key}')
        if prop is not None:
            processed_data[key] = commands.cast_backend_to_python(
                context,
                prop,
                backend,
                value,
                **kwargs
            )

    for prop in dtype.refprops:
        if prop.name not in processed_data and prop.name in data:
            processed_data[prop.name] = commands.cast_backend_to_python(
                context,
                prop,
                backend,
                data[prop.name],
                **kwargs
            )

    if not processed_data or all(value is None for value in processed_data.values()):
        return None

    return processed_data


@commands.cast_backend_to_python.register(Context, Ref, Sql, object)
def cast_backend_to_python(
    context: Context,
    dtype: Ref,
    backend: Sql,
    data: object,
    **kwargs
):
    if len(dtype.refprops) != 1:
        raise GivenValueCountMissmatch(dtype, given_count=1, expected_count=len(dtype.refprops))

    return commands.cast_backend_to_python(
        context,
        dtype,
        backend, {'_id': {
            dtype.refprops[0].name: data
        }},
        **kwargs
    )


@commands.cast_backend_to_python.register(Context, ExternalRef, Sql, object)
def cast_backend_to_python(
    context: Context,
    dtype: Ref,
    backend: Sql,
    data: object,
    **kwargs
):
    if len(dtype.refprops) != 1:
        raise GivenValueCountMissmatch(dtype, given_count=1, expected_count=len(dtype.refprops))

    return commands.cast_backend_to_python(
        context,
        dtype,
        backend, {
            dtype.refprops[0].name: data
        },
        **kwargs
    )


@commands.cast_backend_to_python.register(Context, Ref, Sql, (list, tuple))
def cast_backend_to_python(
    context: Context,
    dtype: Ref,
    backend: Sql,
    data: list | tuple,
    **kwargs
):
    # This kinda of convertion is not preferred (dict values are better), but it's kept to support
    # dialects that cannot cast to json objects.
    # It will only attempt to map dtype.refprops to given values, meaning Denorm mapping is lost
    if len(data) != len(dtype.refprops):
        raise GivenValueCountMissmatch(dtype, given_count=len(data), expected_count=len(dtype.refprops))

    return commands.cast_backend_to_python(
        context,
        dtype,
        backend, {
            '_id': {
                prop.name: data[i] for i, prop in enumerate(dtype.refprops)
            },
        },
        **kwargs
    )


@commands.cast_backend_to_python.register(Context, ExternalRef, Sql, (list, tuple))
def cast_backend_to_python(
    context: Context,
    dtype: ExternalRef,
    backend: Sql,
    data: list | tuple,
    **kwargs
):
    # This kinda of convertion is not preferred (dict values are better), but it's kept to support
    # dialects that cannot cast to json objects.
    # It will only attempt to map dtype.refprops to given values, meaning Denorm mapping is lost
    if len(data) != len(dtype.refprops):
        raise GivenValueCountMissmatch(dtype, given_count=len(data), expected_count=len(dtype.refprops))

    return commands.cast_backend_to_python(
        context,
        dtype,
        backend, {prop.name: data[i] for i, prop in enumerate(dtype.refprops)},
        **kwargs
    )



from typing import List

import geoalchemy2.types
import sqlalchemy as sa

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers.migrate.migrate import MigratePostgresMeta, \
    adjust_kwargs, extract_literal_name_from_column, handle_unique_constraint_migration, contains_unique_constraint, \
    handle_index_migration, extract_using_from_columns, MigrateModelMeta, contains_constraint_name, \
    constraint_with_columns, extract_sqlalchemy_columns, reduce_columns
from spinta.backends.postgresql.helpers.name import name_changed, get_pg_table_name, get_pg_constraint_name, \
    get_pg_removed_name, get_pg_index_name
from spinta.components import Context
from spinta.types.datatype import DataType
from spinta.utils.itertools import ensure_list
from spinta.utils.schema import NotAvailable, NA


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, DataType)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: DataType, **kwargs):
    columns = commands.prepare(context, backend, new.prop)
    columns = ensure_list(columns)
    columns = extract_sqlalchemy_columns(columns)
    columns = reduce_columns(columns)
    if columns is not None and columns != []:
        commands.migrate(context, backend, meta, table, old, columns, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, DataType)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: DataType, **kwargs):
    columns = commands.prepare(context, backend, new.prop)
    columns = ensure_list(columns)
    columns = extract_sqlalchemy_columns(columns)
    columns = reduce_columns(columns)
    commands.migrate(context, backend, meta, table, old, columns, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, DataType)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: List[sa.Column], new: DataType, **kwargs):
    rename = meta.rename

    column: sa.Column = commands.prepare(context, backend, new.prop)
    old_name = rename.get_old_column_name(table.name, column.name)

    columns = old.copy()

    if name_changed(column.name, old_name):
        for col in old:
            if col.name == column.name:
                commands.migrate(context, backend, meta, table, col, NA, **kwargs)
                columns.remove(col)
                break

    for col in columns:
        commands.migrate(context, backend, meta, table, col, column, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, sa.Column)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: sa.Column, model_meta: MigrateModelMeta, foreign_key: bool = False, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    column_name = new.name
    table_name = get_pg_table_name(rename.get_table_name(table.name))
    old_type = extract_literal_name_from_column(old)
    new_type = extract_literal_name_from_column(new)

    nullable = new.nullable if new.nullable != old.nullable else None
    type_ = new.type if old_type != new_type else None
    new_name = column_name if old.name != new.name else None

    using = extract_using_from_columns(
        old,
        new,
        type_
    )

    if nullable is not None or type_ is not None or new_name is not None or using is not None:
        handler.add_action(ma.AlterColumnMigrationAction(
            table_name=table_name,
            column_name=old.name,
            nullable=nullable,
            type_=type_,
            new_column_name=new_name,
            using=using,
        ), foreign_key)

    is_renamed = name_changed(table.name, table_name, old.name, new.name)
    # Order has to be UniqueConstraint -> Index
    # because UniqueConstraint also contain Unique Index
    # we mark them as handled if they are part of UniqueConstraint
    handle_unique_constraint_migration(
        table,
        table_name,
        old,
        new,
        column_name,
        handler,
        inspector,
        foreign_key,
        is_renamed,
        meta=model_meta
    )
    handle_index_migration(
        table,
        table_name,
        old,
        new,
        column_name,
        handler,
        inspector,
        foreign_key,
        is_renamed,
        meta=model_meta
    )


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, sa.Column)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: sa.Column, model_meta: MigrateModelMeta, foreign_key: bool = False, **kwargs):
    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    table_name = get_pg_table_name(rename.get_table_name(table.name))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=new,
    ), foreign_key)

    new_column_name = new.name
    if new.unique:
        constraint_name = get_pg_constraint_name(table_name, [new_column_name])
        unique_constraints = inspector.get_unique_constraints(table_name=table_name)
        model_meta.handle_unique_constraint(constraint_name)
        if not contains_unique_constraint(unique_constraints, new_column_name):
            handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                constraint_name=constraint_name,
                table_name=table_name,
                columns=[new_column_name]
            ))
        elif not contains_constraint_name(unique_constraints, constraint_name):
            constraint = constraint_with_columns(unique_constraints, [new_column_name])
            if constraint:
                model_meta.handle_unique_constraint(constraint['name'])
                handler.add_action(ma.RenameConstraintMigrationAction(
                    table_name=table_name,
                    old_constraint_name=constraint['name'],
                    new_constraint_name=constraint_name
                ))

    index_required = isinstance(new.type, geoalchemy2.types.Geometry)
    if index_required:
        index_name = get_pg_index_name(table_name, new_column_name)
        model_meta.handle_index(index_name)
        handler.add_action(ma.CreateIndexMigrationAction(
            table_name=table_name,
            columns=[new_column_name],
            index_name=index_name,
            using='GIST'
        ))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, NotAvailable)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: list, new: NotAvailable, foreign_key: bool = False, **kwargs):
    for item in old:
        commands.migrate(context, backend, meta, table, item, new, **adjust_kwargs(kwargs, {
            'foreign_key': foreign_key
        }))


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, NotAvailable)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: NotAvailable, model_meta: MigrateModelMeta, foreign_key: bool = False, **kwargs):
    if old.name.startswith("_"):
        return

    rename = meta.rename
    inspector = meta.inspector
    handler = meta.handler

    table_name = get_pg_table_name(rename.get_table_name(table.name))
    columns = inspector.get_columns(table.name)
    remove_name = get_pg_removed_name(old.name)

    if any(remove_name == column["name"] for column in columns):
        handler.add_action(ma.DropColumnMigrationAction(
            table_name=table_name,
            column_name=remove_name
        ), foreign_key)
    handler.add_action(ma.AlterColumnMigrationAction(
        table_name=table_name,
        column_name=old.name,
        new_column_name=remove_name
    ), foreign_key)
    indexes = inspector.get_indexes(table_name=table.name)
    for index in indexes:
        if index["column_names"] == [old.name]:
            model_meta.handle_index(index["name"])
            handler.add_action(ma.DropIndexMigrationAction(
                table_name=table_name,
                index_name=index["name"],
            ), foreign_key)
    if old.unique:
        unique_constraints = inspector.get_unique_constraints(table_name=table.name)
        for constraint in unique_constraints:
            if old.name in constraint["column_names"]:
                model_meta.handle_unique_constraint(constraint["name"])
                handler.add_action(ma.DropConstraintMigrationAction(
                    table_name=table_name,
                    constraint_name=constraint["name"],
                ), foreign_key)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, list)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table, old: NotAvailable,
            new: List[sa.Column], foreign_key: bool = False, **kwargs):
    for column in new:
        commands.migrate(context, backend, meta, table, old, column, foreign_key=foreign_key, **kwargs)



from __future__ import annotations

import contextlib
import warnings

from typing import TYPE_CHECKING

from spinta.core.ufuncs import asttoexpr
from spinta.datasets.backends.dataframe.backends.csv.components import Csv
from spinta.datasets.backends.dataframe.backends.json.components import Json
from spinta.datasets.backends.dataframe.backends.xml.components import Xml

if TYPE_CHECKING:
    from spinta.components import UrlParams
    from spinta.core.ufuncs import Expr


def urlparams_to_expr(
    params: UrlParams,
) -> Expr:
    """Convert UrlParams to ufunc Expr.

    This is compatibility function. Currently all query handling is done using
    UrlParams, but in future UrlParams should be replaced by ufunc Expr. This
    way getall and getone commands will receive just one Expr parameter
    containing all the query information.
    """

    ast = []

    if params.query:
        ast += params.query

    if params.select:
        ast.append({'name': 'select', 'args': [
            arg if isinstance(arg, dict) else {'name': 'bind', 'args': [arg]}
            for arg in params.select
        ]})

    if params.sort:
        ast.append({'name': 'sort', 'args': [
            arg if isinstance(arg, dict) else {'name': 'bind', 'args': [arg]}
            for arg in params.sort
        ]})

    if params.limit:
        ast.append({'name': 'limit', 'args': [params.limit]})

    if params.offset:
        ast.append({'name': 'offset', 'args': [params.offset]})

    if params.bbox is not None:
        ast.append({'name': 'bbox', 'args': params.bbox})

    if len(ast) == 0:
        ast = {
            'name': 'select',
            'args': [],
        }
    elif len(ast) == 1:
        ast = ast[0]
    else:
        ast = {'name': 'and', 'args': ast}

    return asttoexpr(ast)


# Backwards compatibility `xml` backend class.
# It will eventually be deprecated fully.
class XmlDeprecated(Xml):
    type: str = 'xml'

    @contextlib.contextmanager
    def begin(self):
        yield

    def __init__(self):
        super().__init__()
        warnings.warn("'xml' backend type is deprecated, use 'dask/xml'.", FutureWarning)


# Backwards compatibility `json` backend class.
# It will eventually be deprecated fully.
class JsonDeprecated(Json):
    type: str = 'json'

    @contextlib.contextmanager
    def begin(self):
        yield

    def __init__(self):
        super().__init__()
        warnings.warn("'json' backend type is deprecated, use 'dask/json'.", FutureWarning)


# Backwards compatibility `csv` backend class.
# It will eventually be deprecated fully.
class CsvDeprecated(Csv):
    type: str = 'csv'

    @contextlib.contextmanager
    def begin(self):
        yield

    def __init__(self):
        super().__init__()
        warnings.warn("'csv' backend type is deprecated, use 'dask/csv'.", FutureWarning)



from typing import Any, TypedDict
from typing import Dict
from typing import NamedTuple
from typing import Optional

import sqlalchemy as sa

from spinta.components import Model


class State(NamedTuple):
    engine: sa.engine.Engine
    metadata: sa.MetaData


class PushRow:
    model: Optional[Model]
    data: Dict[str, Any]
    # SHA1 checksum of data generated with _get_data_rev.
    checksum: Optional[str]
    # True if data has already been sent.
    saved: bool = False
    # If push request received an error
    error: bool = False
    # Row operation
    op: Optional[str] = None
    # If True, need to push data immediately
    push: bool = False
    # If True, need to include in request
    send: bool = True

    def __init__(
        self,
        model: Optional[Model],
        data: Dict[str, Any],
        checksum: str = None,
        saved: bool = False,
        error: bool = False,
        op: str = None,
        push: bool = False,
        send: bool = True,
    ):
        self.model = model
        self.data = data
        self.checksum = checksum
        self.saved = saved
        self.error = error
        self.op = op
        self.push = push
        self.send = send


class ErrorContext(TypedDict):
    # Manifest name
    manifest: str

    # Dataset name
    dataset: str

    # Absolute model name (but does not start with /)
    model: str

    # Property name
    property: str

    # Property data type
    type: str

    # Model name in data source
    entity: str

    # Property name in data source
    attribute: str

    # Full class path of component involved in the error.
    component: str

    # Line number or other location (depends on manifest) node in manifest
    # involved in the error.
    schema: str

    # External object id
    id: str


class Error(TypedDict):
    # Error type
    type: str

    # Error code (exception class name)
    code: str

    # Error context.
    context: ErrorContext

    # Message template.
    tempalte: str

    # Message compiled from templates and context.
    message: str


class Saved(NamedTuple):
    revision: str
    checksum: str


PUSH_NOW = PushRow(
    model=None,
    data={
        '_type': None
    },
    push=True,
    send=False,
)



from typing import Optional, Any, Dict, Iterable, Tuple

import logging
import re


log = logging.getLogger(__name__)


class UnknownValue:

    def __str__(self):
        return '[UNKNOWN]'

    __repr__ = __str__


UNKNOWN_VALUE = UnknownValue()


def resolve_context_vars(schema: Dict[str, str], this: Optional[Any], kwargs: dict):
    """Resolve value from given kwargs and schema."""
    # Extend context by calling get_error_context command on first positional
    # argument of Node or Type type.
    if this:
        from spinta import commands
        schema = {
            **commands.get_error_context(this),
            **schema,
        }
        kwargs = {**kwargs, 'this': this}

    added = set()
    context = {}
    if this:
        context['component'] = type(this).__module__ + '.' + type(this).__name__
    for k, path in schema.items():
        path = path or k
        name, *names = path.split('.')
        if name not in kwargs:
            continue
        added.add(name)
        value = kwargs
        for name in [name] + names:
            if name.endswith('()'):
                name = name[:-2]
                func = True
            else:
                func = False
            if isinstance(value, dict):
                value = value.get(name)
            elif hasattr(value, name):
                value = getattr(value, name)
            else:
                value = UNKNOWN_VALUE
                break
            if func:
                value = value()
        if value is not UNKNOWN_VALUE:
            context[k] = value

    for k in set(kwargs) - added:
        v = kwargs[k]
        if not isinstance(v, (int, float, str)):
            v = str(v)
        context[k] = v

    # Return sorted context.
    names = [
        'component',
        'manifest',
        'schema',
        'backend',
        'dataset',
        'resource',
        'model',
        'entity',
        'property',
        'attribute',
        'type',
    ]
    names += [x for x in schema if x not in names]
    names += [x for x in kwargs if x not in names]

    def sort_key(item: Tuple[str, Any]) -> Tuple[int, str]:
        key = item[0]
        try:
            return names.index(key), key
        except ValueError:
            return len(names), key

    return {k: v for k, v in sorted(context.items(), key=sort_key)}


class BaseError(Exception):
    type: str = None
    status_code: int = 500
    template: str = None
    headers: Dict[str, str] = {}
    context: Dict[str, Any] = {}

    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            this = None
        elif len(args) == 1:
            this = args[0]
        else:
            this = None
            log.error("Only one positional argument is allowed, but %d was given.", len(args), stack_info=True)

        self.type = this.type if this and hasattr(this, 'type') else 'system'

        self.context = resolve_context_vars(self.context, this, kwargs)

    def __str__(self):
        return (
            self.message + '\n' +
            ('  Context:\n' if self.context else '') +
            ''.join(
                f'    {k}: {v}\n'
                for k, v in self.context.items()
            )
        )

    @property
    def message(self):
        try:
            return _render_template(self)
        except KeyError:
            log.exception("Can't render error message for %s.", self.__class__.__name__)
            return self.template


def error_response(error: BaseError):
    return {
        'type': error.type,
        'code': type(error).__name__,
        'template': error.template,
        'context': error.context,
        'message': error.message,
    }


def _render_template(error: BaseError):
    if error.type in error.context:
        context = {
            **error.context,
            'this': f'<{error.type} name={error.context[error.type]!r}>',
        }
    else:
        context = error.context
    try:
        return error.template.format(**context)
    except KeyError:
        context = context.copy()
        template_vars_re = re.compile(r'\{(\w+)')
        for match in template_vars_re.finditer(error.template):
            name = match.group(1)
            if name not in context:
                context[name] = UNKNOWN_VALUE
        return error.template.format(**context)


class MultipleErrors(Exception):

    def __init__(self, errors: Iterable[BaseError]):
        self.errors = list(errors)
        super().__init__(
            'Multiple errors:\n' + ''.join([
                f' - {error.message}\n' +
                '     Context:\n' + ''.join(
                    f'       {k}: {v}\n' for k, v in error.context.items()
                )
                for error in self.errors
            ])
        )


class UserError(BaseError):
    status_code = 400


class UpgradeError(BaseError):
    status_code = 500


class ConflictingValue(UserError):
    status_code = 409
    template = "Conflicting value."
    context = {
        'given': None,
        'expected': None,
    }


class UniqueConstraint(UserError):
    template = "Given value already exists."


class InvalidOperandValue(UserError):
    template = "Invalid operand value for {operator!r} operator."


class UnknownOperator(UserError):
    template = "Unknown operator {operator!r}."


class ItemDoesNotExist(UserError):
    status_code = 404
    template = "Resource {id!r} not found."


class DatasetNotFound(UserError):
    status_code = 404
    template = "Dataset {dataset!r} not found."


class NamespaceNotFound(UserError):
    status_code = 404
    template = "Namespace {namespace!r} not found."


class ModelNotFound(UserError):
    status_code = 404
    template = "Model {model!r} not found."


class NoModelDefined(UserError):
    template = "Property {property!r} must be defined on a concrete model."


class PropertyNotFound(UserError):
    status_code = 404
    template = "Property {property!r} not found."


class NoItemRevision(UserError):
    template = "'_revision' must be given on rewrite operation."


class JSONError(UserError):
    template = "Invalid JSON."


class InvalidValue(UserError):
    template = "Invalid value."


class InvalidPropertyType(UserError):
    template = "Invalid property type, expected {expected}, got {type}."


class UndefinedPropertyType(UserError):
    template = 'Parameter "type" must be defined for property "{property}", because it is not defined in base model or there is no base model.'


class ValueNotInEnum(UserError):
    template = "Given value {value} is not defined in enum."


class UndefinedEnum(UserError):
    template = "Enum {name!r} is not defined."


class EmptyStringSearch(UserError):
    template = \
        "Empty string can't be used with `{op}`. " \
        "Use `exact` parameter."


class InvalidToken(UserError):
    status_code = 401
    template = "Invalid token"
    headers = {'WWW-Authenticate': 'Bearer error="invalid_token"'}


class BasicAuthRequired(UserError):
    status_code = 401
    template = "Unauthorized"
    headers = {'WWW-Authenticate': 'Basic realm="Authentication required."'}


class AuthorizedClientsOnly(UserError):
    status_code = 401
    template = "This resource can only be accessed by an authorized client."


class MultipleRowsFound(BaseError):
    template = "Multiple rows were found."


class ManagedProperty(UserError):
    template = "Value of this property is managed automatically and cannot be set manually."


class InvalidManagedPropertyName(UserError):
    template = "Invalid managed property name. Expected name {name!r}, got {property!r}."


class NotFoundError(BaseError):
    template = "No results where found."
    status_code = 500


class NodeNotFound(UserError):
    template = "Node {name!r} of type {type!r} not found."


class ModelReferenceNotFound(BaseError):
    template = "Model reference {ref!r} not found."


class ModelReferenceKeyNotFound(BaseError):
    template = "Model reference key {ref!r} not found in {model!r}."


class SourceNotSet(UserError):
    status_code = 404
    template = (
        "Dataset {dataset!r} resource {resource!r} source is not set. "
        "Make sure {resource!r} name parameter in {path} or environment variable {envvar} is set."
    )


class InvalidManifestFile(BaseError):
    template = "Error while parsing {eid!r} manifest entry: {error}"


class CoordinatesOutOfRange(UserError):
    template = 'Given coordinates: {given!r} ar not within the `EPSG: {srid!r}` available bounds: {bounds} (west, south, east, north).'


class ManifestFileDoesNotExist(BaseError):
    template = "Manifest file {path} does not exist."


class UnknownProjectOwner(BaseError):
    template = "Unknown owner {owner}."
    context = {
        'owner': 'this.owner'
    }


class UnknownProjectDataset(BaseError):
    template = "Unknown project dataset."
    context = {
        'manifest': 'project.parent.name',
        'dataset': 'project.dataset',
        'project': 'project.name',
        'filename': 'project.path',
    }


class MissingRequiredProperty(BaseError):
    template = "Property {prop!r} is required."


class FileNotFound(UserError):
    template = "File {file!r} not found."


class UnknownModelReference(BaseError):
    template = "Unknown model reference given in {param!r} parameter."


class InvalidDependencyValue(BaseError):
    template = "Dependency must be in 'object/name.property' form, got: {dependency!r}."


class MultipleModelsInDependencies(BaseError):
    template = "Dependencies are allowed only from single model, but more than one model found: {models}."


class MultipleCommandCallsInDependencies(BaseError):
    template = "Only one command call is allowed."


class MultipleCallsOrModelsInDependencies(BaseError):
    template = "Only one command call or one model is allowed in dependencies."


class InvalidSource(BaseError):
    template = "Invalid source. {error}"
    context = {
        'source': 'source.type',
    }


class UnknownParameter(BaseError):
    template = "Unknown parameter {parameter!r}."
    context = {
        'parameter': 'param',
    }


class InvalidUuidValue(UserError):
    template = "Invalid uuid value: {value}."


class InvalidRefValue(UserError):
    template = "Invalid reference value: {value}."


class InvalidLevel(UserError):
    template = "Invalid level value \"{level}\"."


class InvalidParameterValue(BaseError):
    template = "Invalid parameter {parameter!r} value."


class TooManyParameters(BaseError):
    template = (
        "Too many parameters, you can only give up to {max_params} parameters."
    )


class FieldNotInResource(UserError):
    template = "Unknown property {property!r}."
    context = {
        'property': 'prop',
    }


class UnknownContentType(UserError):
    status_code = 415
    template = "Unknown content type {content_type!r}."


class UnknownAction(UserError):
    template = "Unknown action {action!r}."


class UnknownOutputFormat(UserError):
    template = "Unknown output format {name!r}."


class UnknownRequestParameter(UserError):
    template = "Unknown request parameter {name!r}."


class OutOfScope(UserError):
    template = "{this} is out of given scope {scope!r}."
    context = {
        'scope': 'scope.name',
    }


class UnhandledException(BaseError):
    status_code = 500
    template = "Unhandled exception {exception}: {error}."
    context = {
        'exception': 'error.__class__.__name__',
    }


class NewItemHasIdAlready(UserError):
    template = "New item has id already set"


class NewItemHasRevisionAlready(UserError):
    template = "New item has revision already set"


class ModelSourceNotFound(UserError):
    template = "Model source {table!r} not found."


class UnavailableSubresource(UserError):
    template = "Subresources only of type Object and File are accessible"


class InvalidPostUrl(UserError):
    template = "Invalid URL for POST request"


class MultipleParentsError(Exception):
    pass


class UnknownPropertyType(UserError):
    template = "Unknown property type {type!r}."


class UnknownMethod(UserError):
    template = "Unknown method {name!r} with args {expr}."


class MissingReference(UserError):
    template = "Missing reference {ref!r} referenced from {param!r} parameter."


class UnacceptableFileName(UserError):
    template = "Path is not acceptable in filename {file!r}"


class NoAuthServer(UserError):
    template = "Authorization server is disabled, use an external authorization server."


class NoTokenValidationKey(UserError):
    template = "A token validation key is required, set it via token_validation_key configuration parameter."


class NoExternalName(UserError):
    template = "Property {property} does not have 'external' name."


class NoKeyMap(UserError):
    template = "Key map is not configured."


class UnknownKeyMap(UserError):
    template = "Keymap {keymap!r} is not found."


class BackendNotFound(UserError):
    template = "Can't find backend {name!r}."


class NoBackendConfigured(UserError):
    template = "Backend is not configured, can't proceed the request."


class UnexpectedFormulaResult(UserError):
    template = (
        "Unexpected formula {formula} result. "
        "Expected {expected}, instead got a {received}."
    )


class FormulaError(UserError):
    template = "Error while interpreting formula {formula}: {error}."


class UnknownBind(FormulaError):
    template = "Unknown bind {name!r}."


class RequiredConfigParam(UserError):
    template = "Configuration parameter {name!r} is required."


class IncompatibleForeignProperties(UserError):
    template = (
        "Can't join {this} and {right}, these two properties does not have "
        "direct connection with one another."
    )


class KeymapNotSet(UserError):
    template = (
        "Keymap is required for {this}, but is not configured. Please make "
        "sure a key map is configured."
    )


class NoRefPropertyForDenormProperty(UserError):
    template = (
        "Property {ref!r} with type 'ref' or 'object' must be defined "
        "before defining property {prop!r}."
    )


class ReferencedPropertyNotFound(PropertyNotFound):
    pass


class RemoteClientError(UserError):
    pass


class RemoteClientCredentialsNotFound(RemoteClientError):
    template = (
        "Remote client credentials not found for {name!r}. Make sure, a "
        "section [{section}] exists in {credentials} file. You can use "
        "`spinta remote add {name}` command to add it."
    )


class RemoteClientCredentialsNotGiven(RemoteClientError):
    template = (
        "Make sure client name and secret is given in {credentials} file, "
        "[{section}] section."
    )


class RemoteClientScopesNotGiven(RemoteClientError):
    template = (
        "Make sure at least one scope is given for [{section}] in "
        "{credentials} file."
    )


class DupicateProperty(UserError):
    template = "Duplicate property {name}."


class RequiredProperty(UserError):
    template = "Property is required."


class UnableToCast(UserError):
    template = "Unable to cast {value} to {type} type."


class NotImplementedFeature(BaseError):
    template = "{feature} is not implemented yet."


class ReferencedObjectNotFound(UserError):
    template = "Referenced object {id!r} not found."


class ReferringObjectFound(UserError):
    template = "Object {id!r} is still being referenced in table {model!r}."


class CompositeUniqueConstraint(UserError):
    template = "Given values for composition of properties ({properties}) already exist."


class SourceCannotBeList(BaseError):
    template = "Source can't be a list, use prepare instead."


class InsufficientPermission(UserError):
    status_code = 403
    template = "You need to have {scope!r} in order to access this API endpoint."


class InsufficientPermissionForUpdate(UserError):
    status_code = 403
    template = "You do not have a permission to update '{field}' field."


class UnknownPropertyInRequest(UserError):
    template = "Property '{property}' is not part of allowed properties: '{properties}'"


class ClientWithNameAlreadyExists(UserError):
    template = "Client with name '{client_name}' already exists."


class ClientAlreadyExists(UserError):
    template = "Client '{client_id}' already exists."


class EmptyPassword(UserError):
    template = "Client password cannot be empty."


class UnknownRequestQuery(UserError):
    template = "Request '{request}' does not support '{query}' query."


class InvalidRequestQuery(UserError):
    template = "Query '{query}' requires '{format}' format."


class UnexpectedFormKeys(UserError):
    template = "Unexpected keys: {unknown_keys}, only keys that are allowed are: {allowed_keys}."


class RequiredFormKeyWithCondition(UserError):
    template = "'{key}' key is required when {condition}."


class InvalidFormKeyCombination(UserError):
    template = "Form only accepts one of the keys: {keys}."


class MissingFormKeys(UserError):
    template = "Form requires to have at least one of the keys: {keys}."


class InvalidInputData(UserError):
    template = "'{key}' does not accept value: '{given}', {condition}"


class InvalidResourceSource(UserError):
    template = "'{source}' is unacceptable resource source, it must be URL."


class UnknownManifestType(BaseError):
    template = "Can't find manifest component matching given type {type!r}."


class UnknownManifestTypeFromPath(BaseError):
    template = "Can't find manifest type matching given path {path!r}."


class InvalidPageParameterCount(UserError):
    template = "Parameter 'page' only accepts one of page key, size, or disable attributes."


class InvalidPushWithPageParameterCount(UserError):
    template = "Given push page property count does not match model required properties: {properties}."


class InvalidPageKey(UserError):
    template = "Given '{key}' page key is invalid."


class InfiniteLoopWithPagination(UserError):
    template = '''
    Pagination values has cause infinite loop while fetching data.
    Page of size: {page_size}, first value is the same as previous page's last value, which is:
    {page_values}
    '''


class TooShortPageSize(UserError):
    template = '''
    Page of size: {page_size} is too small, some duplicate values do not fit in a single page.
    Which can cause either loss of data, or cause infinite loop while paginating.
    Affected row: {page_values}
    
    To fix this, please either increase page size in the manifest, or 'default_page_size' value in the configs.
    Alternatively make page's structure more complex, by adding more properties to it.
    
    When migrating from older versions to newer versions of spinta you might get this error if push state
    database is out of sync, add '--sync' tag to 'push' command to synchronize it.
    '''


class TooShortPageSizeKeyRepetition(TooShortPageSize):

    def __init__(self, *args, **kwargs):
        super(TooShortPageSize, self).__init__(*args, **kwargs)
        self.template = f'''
        {self.template}
        Error has been triggered because:
        New page's key has been encountered multiple times in the previous page and it is the same for the future value.
        This will cause the same data to be fetched multiple times.
        '''


class DuplicateRowWhilePaginating(BaseError):
    template = "Encountered a duplicate row with page key: '{key}'"


class UnauthorizedPropertyPush(UserError):
    code = 403
    template = "You do not have permission to push this property."


class InvalidArgumentInExpression(BaseError):
    template = "Invalid {arguments} arguments given to {expr} expression."


class BackendNotGiven(UserError):
    template = "Model is operating in external mode, yet it does not have assigned backend to it."


class UnauthorizedKeymapSync(UserError):
    code = 403
    template = "You do not have permission to sync this model's keymap."


class GivenValueCountMissmatch(BaseError):
    template = '''
    While assigning ref values {given_count} were given, but {expected_count} were expected.
    This can happen, when there are no primary keys set on ref's model, invalid ref key assignment or the keys can be nullable.
    '''


class PartialTypeNotFound(BaseError):
    template = "Partial type can only be used for ref type."


class NoReferencesFound(UserError):
    template = "Property {prop_name!r} of type 'Ref' was not found."


class MultipleBackRefReferencesFound(UserError):
    template = "Model {model!r} contains multiple references to backref, please specify which one to use."


class NoBackRefReferencesFound(UserError):
    template = "Model {model!r} does not contain any suitable properties for backref."


class InvalidBackRefReferenceAmount(UserError):
    template = "Backref {backref!r} can only reference one property."


class CannotModifyBackRefProp(UserError):
    template = "It is impossible to directly set or modify Backref property."


class OneToManyBackRefNotSupported(UserError):
    template = "One to many relationship is not supported by Backref."


class SummaryWithMultipleProperties(UserError):
    template = "Summary with multiple properties is not supported."


class TooManyModelUriProperties(UserError):
    template = "Model already has {uri_prop!r} set as uri property."


class DataTypeCannotBeUsedForNesting(UserError):
    template = "Type {dtype!r} cannot be used for nesting properties."


class NestedDataTypeMismatch(UserError):
    template = "While nesting, {initial!r} type cannot be cast to {required!r} type."


class LangNotDeclared(UserError):
    template = "Language {lang!r} has not been declared."


class TooManyLangsGiven(UserError):
    template = "Too many languages given in 'content-language' header, expected only one, but were given {amount}."


class UnableToDetermineRequiredLang(UserError):
    template = "Unable to determine required language."


class CannotSelectTextAndSpecifiedLang(UserError):
    template = "Cannot select undisclosed language Text at the same time when disclosed language Text is selected."


class DuplicateRdfPrefixMissmatch(UserError):
    template = "Currently system does not support prefix missmatch. Prefix {prefix!r} has {old_value!r} and {new_value!r} values given."


class InvalidName(UserError):
    template = 'Invalid {name!r} {type} code name.'


class NoneValueComparison(UserError):
    template = "None values can only be compared using 'eq' or 'ne' operands, {op!r} was given."


class InvalidDenormProperty(UserError):
    template = 'Cannot create Denorm property {denorm!r}, because it is part of {ref!r} refprops: {refprops}.'


class RefPropTypeMissmatch(UserError):
    template = 'Refprop {refprop!r} requires {required_type!r} type, but was given {given_type!r}.'


class InheritPropertyValueMissmatch(UserError):
    template = 'Expected {expected!r} value, but got {given!r}.'


class OutOfMemoryMigrate(UserError):
    template = "Ran out of shared memory while migrating. Use 'spinta migrate --autocommit' flag, or increase 'max_locks_per_transaction'."


class ManifestObjectNotDefined(UserError):
    template = "Object {obj!r} is not defined in manifest objects list."


class InvalidIdType(UserError):
    template = "Id {id!r} of {id_type!r} type is invalid."


class NotSupportedManifestType(UserError):
    template = "This feature is only supported with {supported_type} manifests, but {manifest_name!r} is a {given_type} manifest."


class InvalidSchemaUrlPath(UserError):
    template = "Cannot directly alter model or namespace schema. Only datasets are allowed."


class ModifyOneDatasetSchema(UserError):
    template = "Schema modification requires 1 dataset to be given, but {given_amount} were given."


class DatasetNameMissmatch(UserError):
    template = "Expected {expected_dataset!r} dataset, but were given {given_dataset!r}."


class DatasetSchemaRequiresIds(UserError):
    template = "All given schema rows require UUID to be set in the id field."


class ModifySchemaRequiresFile(UserError):
    template = "To modify dataset's schema, you are required to attach csv type manifest."


class FileSizeTooLarge(UserError):
    template = "Given file is too large, only up to {allowed_amount} {measure} is allowed."


class SRIDNotSetForGeometry(BaseError):
    template = "Geometry SRID is required, but was given None."


class KeyNotFound(UserError):
    template = "{key!r} key is not in given data dictionary keys: {dict_keys!r}."


class InvalidParamSource(UserError):
    template = "Unable to recognize {param!r} param's source {source!r} type, given: {given_type!}, expected: {expected_types!r}."


class MigrateScalarToRefTooManyKeys(UserError):
    template = '''
    Migration between scalar types and Ref type is only supported when targeted Ref's model contains
    only 1 primary key (except Ref level 3 to scalar), but were given: {primary_keys}
    '''


class MigrateScalarToRefTypeMissmatch(UserError):
    template = '''
    Migration between scalar types and Ref requires, that mapped columns match their types.
    {details}
    '''


class KeyMapGivenKeyMissmatch(UserError):
    template = '''
    The encoding for the {name!r} keymap already includes the key {found_key!r}, associated with the value {value!r}. 
    You attempted to assign the new primary key {given_key!r}, which conflicts with the existing key. 
    Make sure that all keymap values are unique.
    '''


class MultiplePrimaryKeyCandidatesFound(UserError):
    template = '''
    While assigning foreign key for non-primary key `Ref` property, using values: {values!r}
    found multiple possible matches. This can occur when non-primary keys are not unique and there are duplicate values.
    '''


class NoPrimaryKeyCandidatesFound(UserError):
    template = '''
    While assigning foreign key for non-primary key `Ref` property, using values: {values!r}
    no possible matches were found. This can occur when trying to assign values that do not exist in foreign table.
    '''


class ClientsMigrationRequired(UpgradeError):
    template = '''
    Clients folder structure is out of date. Please migrate it using:
    'spinta upgrade', or 'spinta upgrade -r clients' commands.
    
    Old structure used to be:
    ../clients/???.yml
    
    New structure is:
    ../clients/helpers/keymap.yml
    ../clients/id/??/??/???.yml
    
    Where `keymap.yml` stores `client_name` and `client_id` mapping.
    '''


class ClientsKeymapNotFound(UpgradeError):
    template = '''
    Cannot find `../clients/helpers/keymap.yml` file.
    
    Make sure it exists.
    Consider running `spinta upgrade` or `spinta upgrade -r clients` commands 
    '''


class ClientsIdFolderNotFound(UpgradeError):
    template = '''
    Cannot find `../clients/id` folder.

    Make sure it exists.
    Consider running `spinta upgrade` or `spinta upgrade -r clients` commands 
    '''


class InvalidClientsKeymapStructure(UpgradeError):
    template = '''
    Could not load Clients `keymap.yml`.
    Structure is invalid.

    Fix it or consider running `spinta upgrade -f -r clients` command.
    '''


class UpgradeScriptNotFound(UserError):
    template = '''
    Upgrade script {script!r} not found.
    Available scripts: {available_scripts}.
    '''


class InvalidScopes(UserError):
    template = "Request contains invalid, unknown or malformed scopes: {scopes}."


class DirectRefValueUnassignment(UserError):
    template = '''
    Cannot directly set ref's _id value to None.
    You have to set ref to None, which will also remove all additional stored values related to that ref (child properties).
    '''


class BackendUnavailable(BaseError):
    template = '''
    Unable to access {name!r} backend, please try again later.
    '''


class InvalidClientFileFormat(UserError):
    template = "File {client_file} data must be a dictionary, not a {client_file_type}."


class MissingRefModel(UserError):
    template = 'Property "{property_name}" of type "{property_type}" in the model "{model_name}" should have a model name in the `ref` column, to which it refers.'


class UnableToMapIntermediateTable(UserError):
    template = 'Unable to map intermediate table to `Array` property.'


class InvalidIntermediateTableMappingRefCount(UserError):
    template = 'Intermediate table can only be mapped with 2 properties (left and right), but were given {ref_count}.'


class SameModelIntermediateTableMapping(UserError):
    template = '''
    Intermediate table is referencing same model, need to explicitly give the mapping order.
    Set it as intermediate table's model primary key, or by explicitly setting reference properties.
    First is left property (source), second is right property (targeted value).
    '''


class IntermediateTableMappingInvalidType(UserError):
    template = '''
    Intermediate table requires `Ref` type columns for it's mapping.
    {property_name!r} property is of type {property_type!r}.
    '''


class IntermediateTableValueTypeMissmatch(UserError):
    template = '''
    Array item type needs to match intermediate table's right property type.
    Given array type: {array_type!r}, intermediate table's type: {intermediate_type!r}.
    '''


class IntermediateTableRefModelMissmatch(UserError):
    template = '''
    Intermediate table's left property `Ref` model needs to be the same as array's model.
    Array's model: {array_model!r}, intermediate table's left property `Ref` model: {left_model!r}.
    '''


class IntermediateTableRefPropertyModelMissmatch(UserError):
    template = '''
    Array's `Ref` property model does not match intermediate table's right `Ref` property model.
    Array's `Ref` model: {array_ref_model!r}, intermediate table's right property `Ref` model: {right_model!r}.
    '''


class IntermediateTableMissingMappingProperty(UserError):
    template = '''
    Intermediate table's {side} property cannot be None.
    '''


class UnableToFindPrimaryKeysNoUniqueConstraints(UserError):
    template = '''
    Unable to find primary keys for table: {table_name!r}.
    It does not contain any `UniqueConstraints` (primary keys are part of `UniqueConstraint`).
    '''


class UnableToFindPrimaryKeysMultipleUniqueConstraints(UserError):
    template = '''
    Unable to find primary keys for table: {table_name!r}.
    Multiple `UniqueConstrains` were found (that do not match new model's primary keys):
    {unique_constraints}
    '''


class ParentNodeNotFound(UserError):
    template = '''
    Nested properties are defined in '{model_name}' model, without a property that references the parent model.
    Nested properties: '{property_names}'.
    Missing property name: '{missing_property_name}'.
    '''


class PropertyNotPartOfRefProps(UserError):
    template = '''
    {prop!r} is not part of refprops ({refprops})
    '''


class KeymapValueNotFound(UserError):
    template = '''
    KeyMap ({keymap!r}) was unable to get primary key for {model_name!r} model using given values ({values!r}).
    To fix this run `keymap sync` command so that old keys are updated and new ones are added.
    
    This error can occur when trying to get primary keys for internal model, while system is running in external mode and
    keymap does not contain these specific values.
    '''



from typing import List
import sqlalchemy as sa

from spinta import commands
from spinta.auth import get_default_auth_client_id
from spinta.components import Context, Action, Config, Namespace
from spinta.core.enums import Access
from spinta.exceptions import AuthorizedClientsOnly
from spinta.manifests.internal_sql.components import InternalSQLManifest
from spinta.utils.enums import get_enum_by_name
from spinta.utils.scopes import name_to_scope


def get_transaction_connection(context: Context):
    if context.has('transaction.manifest'):
        return context.get('transaction.manifest').connection
    return None


def get_namespace_highest_access(context: Context, manifest: InternalSQLManifest, namespace: str):
    conn = get_transaction_connection(context)
    highest = None
    if conn is not None:
        table = manifest.table
        results = conn.execute(sa.select(table.c.access, sa.func.min(table.c.mpath).label('mpath')).where(
            sa.and_(
                table.c.mpath.startswith(namespace),
                sa.or_(
                    table.c.dim == 'ns',
                    table.c.dim == 'dataset',
                    table.c.dim == 'model',
                    table.c.dim == 'property'
                ),
            )
        ).group_by(table.c.access))

        for result in results:
            if result['access'] is not None:
                enum = get_enum_by_name(Access, result['access'])
                if highest is None or enum > highest:
                    highest = enum
    else:
        objs = manifest.get_objects()
        if namespace in objs['ns']:
            highest = objs['ns'][namespace].access
    return highest if highest is not None else manifest.access


def internal_authorized(
    context: Context,
    name: str,
    access: Access,
    action: Action,
    parents: List[str],
    *,
    throw: bool = False,
):
    config: Config = context.get('config')
    token = context.get('auth.token')

    # Unauthorized clients can only access open nodes.
    unauthorized = token.get_client_id() == get_default_auth_client_id(context)

    open_node = access >= Access.open
    if unauthorized and not open_node:
        if throw:
            raise AuthorizedClientsOnly()
        else:
            return False

    # Private nodes can only be accessed with explicit node scope.
    scopes = [name]

    # Protected and higher level nodes can be accessed with parent nodes scopes.
    if access > Access.private:
        scopes.extend(parents)

    if not isinstance(action, (list, tuple)):
        action = [action]
    scopes = [
        internal_scope_formatter(context, scope, act)
        for act in action
        for scope in scopes
    ]

    # Check if client has at least one of required scopes.
    if throw:
        token.check_scope(scopes, operator='OR')
    else:
        return token.valid_scope(scopes, operator='OR')


def internal_scope_formatter(
    context: Context,
    name: str,
    action: Action,
) -> str:
    config = context.get('config')

    return name_to_scope(
        '{prefix}{name}_{action}' if name else '{prefix}{action}',
        name,
        maxlen=config.scope_max_length,
        params={
            'prefix': config.scope_prefix,
            'action': action.value,
        },
    )


@commands.authorize.register(Context, Action, Namespace, InternalSQLManifest)
def authorize(context: Context, action: Action, ns: Namespace, manifest: InternalSQLManifest):
    parents = [parent.name for parent in ns.parents()]
    return internal_authorized(
        context,
        ns.name,
        get_namespace_highest_access(
            context,
            manifest,
            ns.name
        ),
        action,
        parents,
        throw=True
    )



import sqlalchemy as sa

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import ExternalRef, Denorm
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name


@commands.prepare.register(Context, PostgreSQL, ExternalRef)
def prepare(context: Context, backend: PostgreSQL, dtype: ExternalRef, propagate=True, **kwargs):
    columns = []
    if not dtype.inherited:
        if dtype.model.given.pkeys or dtype.explicit:
            for prop in dtype.refprops:
                # Skip if refprop is already defined in properties, but is not Denorm type
                if prop.name in dtype.properties.keys():
                    if not isinstance(dtype.properties[prop.name].dtype, Denorm):
                        continue

                original_place = prop.place
                prop.place = f"{dtype.prop.place}.{prop.place}"
                if dtype.prop.list and dtype.prop.place == dtype.prop.list.place:
                    prop.place = original_place
                column = commands.prepare(
                    context,
                    backend,
                    prop.dtype,
                )
                if isinstance(column, list):
                    columns.extend(column)
                elif column is not None:
                    columns.append(column)
                prop.place = original_place
            if dtype.unique:
                extracted_columns = [column for column in columns if isinstance(column, sa.Column)]
                unique_constraint = sa.UniqueConstraint(
                    *extracted_columns
                )
                columns.append(unique_constraint)

        else:
            column_name = get_column_name(dtype.prop) + '._id'
            if dtype.prop.list and dtype.prop.place == dtype.prop.list.place:
                column_name = '_id'
            column_type = commands.get_primary_key_type(context, backend)
            columns.append(sa.Column(
                column_name,
                column_type,
                nullable=not dtype.required,
                unique=dtype.unique,
            ))

    if propagate:
        for key, value in dtype.properties.items():
            res = commands.prepare(context, backend, value, **kwargs)
            if res is not None:
                if isinstance(res, list):
                    columns.extend(res)
                else:
                    columns.append(res)
    return columns



from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.backends.oracle.ufuncs.query.components import OracleQueryBuilder

import sqlalchemy as sa

from spinta.datasets.backends.sql.backends.helpers import nulls_last_asc, nulls_first_desc


@ufunc.resolver(OracleQueryBuilder, sa.sql.expression.ColumnElement)
def asc(env, column):
    return nulls_last_asc(column)


@ufunc.resolver(OracleQueryBuilder, sa.sql.expression.ColumnElement)
def desc(env, column):
    return nulls_first_desc(column)



from __future__ import annotations

from spinta import commands
from spinta.components import Property
from spinta.types.datatype import Array


@commands.resolve_property.register(Array, str)
def resolve_property(dtype: Array, prop: str) -> Property | None:
    return commands.resolve_property(dtype.items, prop)



from typing import List, Any

from spinta import commands, spyna
from spinta.components import Context, Model
from spinta.core.ufuncs import asttoexpr, Expr
from spinta.datasets.components import Param, Dataset
from spinta.dimensions.param.components import ParamLoader
from spinta.manifests.components import Manifest
from spinta.nodes import load_node
from spinta.utils.schema import NA


def load_params(context: Context, manifest: Manifest, param_data: Any) -> List[Param]:
    params = []
    if isinstance(param_data, dict):
        for key, data in param_data.items():
            param = Param()
            load_node(context, param, data)
            param.sources = data['source'].copy()
            param.formulas = [asttoexpr(prep) for prep in data['prepare']]
            param.name = data['name']
            params.append(param)
    elif isinstance(param_data, list):
        for item in param_data:
            if isinstance(item, dict) and len(item.keys()) == 1:
                param = Param()
                name = list(item.keys())[0]
                prepare = item[name]
                if isinstance(prepare, str):
                    prepare = spyna.unparse(prepare)
                data = {
                    'name': name,
                    'source': [NA],
                    'prepare': [prepare],
                    'title': '',
                    'description': ''
                }
                load_node(context, param, data)
                param.sources = data['source'].copy()
                param.formulas = [asttoexpr(prep) for prep in data['prepare']]
                param.name = data['name']
                params.append(param)
    return params


def link_params(context: Context, manifest: Manifest, params: List[Param], dataset: Dataset = None):
    param_loader = ParamLoader(context)
    param_loader.update(manifest=manifest, dataset=dataset)
    param_loader.resolve(Expr("resolve_param", params))


def finalize_param_link(context, manifest: Manifest):
    datasets = commands.get_datasets(context, manifest)
    for dataset in datasets.values():
        for resource in dataset.resources.values():
            if resource:
                for param in resource.params:
                    for source in param.sources:
                        if isinstance(source, Model):
                            if source.external and source.external.resource:
                                param.dependencies.update(source.external.resource.source_params)




from spinta import commands
from spinta.components import Context
from spinta.exceptions import UndefinedPropertyType
from spinta.types.datatype import Inherit


@commands.check.register(Context, Inherit)
def check(context: Context, dtype: Inherit) -> None:
    property = dtype.prop
    if property.is_reserved():
        return None

    property_model = property.model
    property_name = property.name
    if not (base := property_model.base) or not base.parent.properties.get(property_name):
        raise UndefinedPropertyType(property_model, property=property_name)



from typing import Union

from spinta.components import Property, FuncProperty
from spinta.core.ufuncs import ufunc, Expr, Bind, GetAttr
from spinta.types.datatype import Integer, String, DataType, Denorm
from spinta.ufuncs.requestparamsbuilder.components import RequestParamsBuilder
from spinta.ufuncs.requestparamsbuilder.helpers import disable_params_pagination


@ufunc.resolver(RequestParamsBuilder, Bind, Bind, name='getattr')
def getattr_(
    env: RequestParamsBuilder,
    field: Bind,
    attr: Bind
):
    return GetAttr(field.name, attr)


@ufunc.resolver(RequestParamsBuilder, Bind, GetAttr, name='getattr')
def getattr_(
    env: RequestParamsBuilder,
    obj: Bind,
    attr: GetAttr
):
    return GetAttr(obj.name, attr)


@ufunc.resolver(RequestParamsBuilder, GetAttr, Bind, name='getattr')
def getattr_(
    env: RequestParamsBuilder,
    obj: GetAttr,
    attr: Bind
):
    leaf = env.call('getattr', obj.name, attr)
    return GetAttr(obj.obj, leaf)


@ufunc.resolver(RequestParamsBuilder, Expr)
def _resolve_empty_dtype(
    env: RequestParamsBuilder,
    expr: Expr,
):
    args, kwargs = expr.resolve(env)
    prop = env.resolve_property(*args, **kwargs)
    return env.call('_resolve_empty_dtype', prop)


@ufunc.resolver(RequestParamsBuilder, Property)
def _resolve_empty_dtype(
    env: RequestParamsBuilder,
    prop: Property,
):
    return env.call('_resolve_empty_dtype', prop.dtype)


@ufunc.resolver(RequestParamsBuilder, FuncProperty)
def _resolve_empty_dtype(
    env: RequestParamsBuilder,
    prop: FuncProperty,
):
    return env.call('_resolve_empty_dtype', prop.prop)


@ufunc.resolver(RequestParamsBuilder, DataType)
def _resolve_empty_dtype(
    env: RequestParamsBuilder,
    dtype: DataType,
):
    new_dtype = dtype.__copy__()
    new_dtype.prop = None
    return new_dtype


@ufunc.resolver(RequestParamsBuilder, Denorm)
def _resolve_empty_dtype(
    env: RequestParamsBuilder,
    dtype: Denorm,
):
    return env.call('_resolve_empty_dtype', dtype.rel_prop)


@ufunc.resolver(RequestParamsBuilder, Expr)
def select(env: RequestParamsBuilder, expr: Expr):
    for arg in expr.args:
        result_name = str(arg)
        result = env.resolve(arg)
        env.call('select', result_name, result, given=arg)


@ufunc.resolver(RequestParamsBuilder, str, FuncProperty)
def select(env: RequestParamsBuilder, select_name: str, func_prop: FuncProperty, given: object = None, **kwargs):
    params = env.params

    if func_prop.func is None:
        func_prop.func = given

    if params.select_funcs is None:
        params.select_funcs = {}
    params.select_funcs[select_name] = func_prop

    if params.select_props is None:
        params.select_props = {}


@ufunc.resolver(RequestParamsBuilder, str, (GetAttr, Bind, Expr))
def select(env: RequestParamsBuilder, select_name: str, other: Union[Bind, Expr], **kwargs):
    params = env.params
    if params.select_props is None:
        params.select_props = {}
    params.select_props[select_name] = other

    if params.select_funcs is None:
        params.select_funcs = {}


@ufunc.resolver(RequestParamsBuilder)
def count(env: RequestParamsBuilder):
    disable_params_pagination(env.params)

    prop = Property()
    prop.name = 'count()'
    prop.place = 'count()'
    prop.title = ''
    prop.description = ''
    prop.model = env.params.model
    prop.dtype = Integer()
    prop.dtype.type = 'integer'
    prop.dtype.type_args = []
    prop.dtype.name = 'integer'
    prop.dtype.prop = prop

    return FuncProperty(func=None, prop=prop)


@ufunc.resolver(RequestParamsBuilder, Expr)
def checksum(env: RequestParamsBuilder, expr: Expr):
    name = str(expr)
    prop = Property()
    prop.name = name
    prop.place = name
    prop.title = ''
    prop.description = ''
    prop.model = env.params.model
    prop.dtype = String()
    prop.dtype.type = 'string'
    prop.dtype.type_args = []
    prop.dtype.name = 'string'
    prop.dtype.prop = prop

    return FuncProperty(func=None, prop=prop)


@ufunc.resolver(RequestParamsBuilder, Expr)
def flip(env: RequestParamsBuilder, expr: Expr):
    name = str(expr)
    prop = Property()
    prop.name = name
    prop.place = name
    prop.title = ''
    prop.description = ''
    prop.model = env.params.model

    dtype = env.call('_resolve_empty_dtype', expr)
    dtype.prop = prop

    prop.dtype = dtype

    return FuncProperty(func=None, prop=prop)




import os
import tempfile

from starlette.requests import Request, FormData
from starlette.datastructures import UploadFile

from typing import Tuple

from spinta import commands
from spinta.auth import check_scope, Scopes
from spinta.components import Context, UrlParams
from spinta.datasets.inspect.helpers import create_manifest_from_inspect
from spinta.exceptions import UnexpectedFormKeys, InvalidFormKeyCombination, RequiredFormKeyWithCondition, \
    MissingFormKeys, InvalidInputData
from spinta.manifests.components import ManifestPath, Manifest


def _validate_form_data(form: FormData):
    # Check form key boundary
    allowed_values = [
        'dataset',
        'manifest.type',
        'manifest.file',
        'manifest.source',
        'resource.type',
        'resource.file',
        'resource.source',
        'resource.prepare'
    ]

    if not set(form.keys()).issubset(allowed_values):
        unknown_keys = list(form.keys() - allowed_values)
        raise UnexpectedFormKeys(allowed_keys=allowed_values, unknown_keys=unknown_keys)


class InspectRequestForm:
    dataset: str
    manifest_path: str = None
    manifest_is_file: bool = False
    manifest_type: str = None
    resource_path: str = None
    resource_is_file: bool = False
    resource_type: str = None
    resource_prepare: str = None

    form: FormData = None

    _manifest_tmp_file: tempfile.NamedTemporaryFile = None
    _resource_tmp_file: tempfile.NamedTemporaryFile = None

    def __init__(self, form: FormData):
        _validate_form_data(form)
        self.form = form

    async def init_data(self):
        await self._load_data(
            dataset=self.form.get("dataset", None),
            manifest_type=self.form.get("manifest.type", None),
            manifest_file=self.form.get("manifest.file", None),
            manifest_source=self.form.get("manifest.source", None),
            resource_type=self.form.get("resource.type", None),
            resource_file=self.form.get("resource.file", None),
            resource_source=self.form.get("resource.source", None),
            resource_prepare=self.form.get("resource.prepare", None)
        )

    def clean_up(self):
        if self.resource_path and self.resource_is_file:
            os.unlink(self.resource_path)

        if self.manifest_path and self.manifest_is_file:
            os.unlink(self.manifest_path)

    def get_manifest(self) -> ManifestPath:
        if self.manifest_path:
            _type = self.manifest_type or ManifestPath.type
            return ManifestPath(type=_type, path=self.manifest_path)

    def get_resource(self) -> Tuple:
        if self.resource_path:
            _type = self.resource_type or self.manifest_type or ManifestPath.type
            return _type, self.resource_path

    async def _load_data(
        self,
        dataset: str,
        manifest_type: str = None,
        manifest_file: UploadFile = None,
        manifest_source: str = None,
        resource_file: UploadFile = None,
        resource_source: str = None,
        resource_type: str = None,
        resource_prepare: str = None
    ):
        if manifest_file and manifest_source:
            raise InvalidFormKeyCombination(keys=["manifest.file", "manifest.source"])

        if manifest_source:
            if "http://" not in manifest_source and "https://" not in manifest_source:
                raise InvalidInputData(key="manifest.source", given=manifest_source, condition="it must be URL")
            self.manifest_path = manifest_source
        elif manifest_file:
            with tempfile.NamedTemporaryFile(delete=False) as tmp:
                read = await manifest_file.read()
                tmp.write(read)
                self.manifest_path = tmp.name
                self.manifest_is_file = True

        self.manifest_type = manifest_type
        if not self.manifest_type and self.manifest_path:
            raise RequiredFormKeyWithCondition(
                key='manifest.type',
                condition="'manifest.source' or 'manifest.file' keys are given"
            )

        if resource_file and resource_source:
            raise InvalidFormKeyCombination(keys=["resource.file", "resource.source"])

        if resource_source:
            if "http://" not in resource_source and "https://" not in resource_source:
                raise InvalidInputData(key="resource.source", given=resource_source, condition="it must be URL")
            self.resource_path = resource_source
        elif resource_file:
            with tempfile.NamedTemporaryFile(delete=False) as tmp:
                read = await resource_file.read()
                tmp.write(read)
                self.resource_path = tmp.name
                self.resource_is_file = True

        self.dataset = dataset
        self.resource_prepare = resource_prepare

        self.resource_type = resource_type
        if self.resource_path and not self.resource_type:
            raise RequiredFormKeyWithCondition(
                key="resource.type",
                condition="'resource.source' or 'resource.file' keys are given"
            )
        if self.resource_path and not self.dataset:
            raise RequiredFormKeyWithCondition(
                key="dataset",
                condition="'resource.source' or 'resource.file' keys are given"
            )

        if not self.manifest_path and not self.resource_path:
            raise MissingFormKeys(keys=[
                'resource.source',
                'resource.file',
                'manifest.source',
                'manifest.file'
            ])


async def inspect_api(context: Context, request: Request, params: UrlParams):
    check_scope(context, Scopes.INSPECT)
    if params.format:
        fmt = params.fmt
    else:
        fmt = context.get("config").exporters["csv"]
    form = await request.form()
    inspect_data = InspectRequestForm(form)
    try:
        await inspect_data.init_data()
        context, manifest = create_manifest_from_inspect(
            dataset=inspect_data.dataset,
            manifest=inspect_data.get_manifest(),
            resources=inspect_data.get_resource(),
            formula=inspect_data.resource_prepare,
            only_url=True
        )
        inspect_data.clean_up()
        clean_up_source_for_return(context, manifest)

        return commands.render(
            context,
            manifest,
            fmt,
            action=params.action,
            params=params)
    except Exception as e:
        inspect_data.clean_up()
        raise e


def clean_up_source_for_return(context: Context, manifest: Manifest):
    for dataset in commands.get_datasets(context, manifest).values():
        for resource in dataset.resources.values():
            if resource.external and not ("http://" in resource.external or "https://" in resource.external):
                resource.external = f"https://get.data.gov.lt/{dataset.name}"



import base64
import datetime
from decimal import Decimal
from typing import Any, Union, Tuple
from typing import Dict
from typing import List
from typing import Optional

from geoalchemy2.elements import WKBElement
from lxml import etree
from lxml.etree import Element, QName
from shapely.geometry.base import BaseGeometry
from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta import commands
from spinta.backends.components import SelectTree
from spinta.backends.helpers import get_model_reserved_props
from spinta.backends.helpers import select_model_props
from spinta.components import Action, Namespace, page_in_data
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.dimensions.prefix.components import UriPrefix
from spinta.exceptions import DuplicateRdfPrefixMissmatch
from spinta.formats.components import Format
from spinta.formats.html.helpers import get_model_link
from spinta.formats.rdf.components import Rdf
from spinta.types.datatype import DataType, PageType, Denorm
from spinta.types.datatype import Date
from spinta.types.datatype import DateTime
from spinta.types.datatype import ExternalRef, Array, BackRef, ArrayBackRef, Inherit, JSON, UUID, Object, Binary
from spinta.types.datatype import File
from spinta.types.datatype import Number
from spinta.types.datatype import Ref
from spinta.types.datatype import Time
from spinta.types.geometry.components import Geometry
from spinta.types.text.components import Text
from spinta.utils.encoding import encode_page_values
from spinta.utils.schema import NotAvailable

RDF = "rdf"
PAV = "pav"
XML = "xml"
DESCRIPTION = "Description"

DEFAULT_PREFIXES = {
    RDF: "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    PAV: "http://purl.org/pav/",
    XML: "http://www.w3.org/XML/1998/namespace"
}


def _get_available_prefixes(context: Context, model: Model) -> dict:
    prefixes = DEFAULT_PREFIXES.copy()
    if commands.has_dataset(context, model.manifest, model.ns.name):
        manifest_prefixes = commands.get_dataset(context, model.manifest, model.ns.name).prefixes
        for key, val in manifest_prefixes.items():
            if isinstance(val, UriPrefix):
                prefixes[key] = val.uri
    return prefixes


def _get_update_prefixes(prefixes: dict, used_datasets: list, model: Model) -> dict:
    if model.external and model.external.dataset and model.external.dataset.name not in used_datasets:
        manifest_prefixes = model.external.dataset.prefixes
        for key, val in manifest_prefixes.items():
            if isinstance(val, UriPrefix):
                if key in prefixes and prefixes[key] != val.uri:
                    raise DuplicateRdfPrefixMissmatch(model.external.dataset, prefix=key, old_value=prefixes[key], new_value=val.uri)
                prefixes[key] = val.uri
        used_datasets.append(model.external.dataset.name)
    return prefixes


def _update_required_prefixes(prefixes: dict) -> dict:
    updated_prefixes = DEFAULT_PREFIXES.copy()
    updated_prefixes.update(prefixes)
    return updated_prefixes


def _get_attribute_name(
    name: str,
    prefix: str,
    available_prefixes: dict
) -> Union[str, QName]:
    if prefix and available_prefixes.get(prefix):
        attribute_name = QName(available_prefixes.get(prefix), name)
    else:
        attribute_name = name
    return attribute_name


def _get_attributes(
    model: Model,
    data: dict,
    about_name: Union[str, QName],
    type_name: Union[str, QName],
    revision_name: Union[str, QName],
) -> Tuple[dict, dict]:
    attributes = {}
    uri_prop = model.uri_prop
    about_uri = False
    if uri_prop is not None:
        if uri_prop.name in data:
            uri = data.pop(uri_prop.name)
            if uri is not None and uri.text:
                attributes[about_name] = uri.text
                about_uri = True
    if '_id' in data:
        _id = data.pop('_id')
        if not about_uri and _id is not None and _id.text:
            attributes[about_name] = get_model_link(model, pk=_id.text)
    if '_type' in data:
        _type = data.pop('_type')
        if _type is not None and _type.text:
            attributes[type_name] = _type.text
    if '_revision' in data:
        _revision = data.pop('_revision')
        if _revision is not None and _revision.text:
            attributes[revision_name] = _revision.text
    return data, attributes


def _get_prefix_and_name(uri: str) -> Tuple[str, str]:
    prefix = name = None
    if uri and ':' in uri:
        prefix = uri.split(':')[0]
        name = uri.split(':')[1]
    elif uri:
        name = uri
    return prefix, name


def _create_element(
    name: Union[str, QName],
    text: str = None,
    base: str = None,
    nsmap: dict = None,
    attributes: dict = None,
    children: List[Element] = None,
) -> Element:
    if attributes is None:
        attributes = {}
    if children is None:
        children = []

    if nsmap is not None:
        elem = Element(name, nsmap=nsmap)
    else:
        elem = Element(name)

    if base is not None:
        elem.base = base
    if text is not None:
        elem.text = text
    for key, value in attributes.items():
        elem.attrib[key] = value
    for child in children:
        if child is not None:
            if isinstance(child, list):
                for item in child:
                    if item is not None:
                        elem.append(item)
            else:
                elem.append(child)
    return elem


def _create_model_element(
    model: Model,
    prefixes: dict,
    data: dict
) -> Element:
    about_name = _get_attribute_name('about', RDF, prefixes)
    type_name = _get_attribute_name('type', RDF, prefixes)
    revision_name = _get_attribute_name('version', PAV, prefixes)
    model_prefix, model_name = _get_prefix_and_name(model.uri)

    data, attributes = _get_attributes(
        model,
        data,
        about_name,
        type_name,
        revision_name
    )
    if model_name:
        name = _get_attribute_name(model_name, model_prefix, prefixes)
    else:
        name = _get_attribute_name(model.basename, model_prefix, prefixes)
    if not isinstance(name, QName):
        name = _get_attribute_name(DESCRIPTION, RDF, prefixes)
        attributes[type_name] = model.model_type()

    return _create_element(
        name=name,
        attributes=attributes,
        children=list(data.values()),
        nsmap=prefixes
    )


def _prepare_for_print(
    model: Model,
    prefixes: dict,
    data: dict
) -> str:
    elem = _create_model_element(model, prefixes, data)
    res = etree.tostring(elem, encoding="unicode", pretty_print=True)
    for key, val in elem.nsmap.items():
        res = res.replace(f'xmlns:{key}="{val}" ', "")
    return res


@commands.render.register(Context, Request, Model, Rdf)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Rdf,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{model.basename}.rdf"'

    return StreamingResponse(
        _stream(
            context,
            request,
            model,
            action,
            data
        ),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


@commands.render.register(Context, Request, Namespace, Rdf)
def render(
    context: Context,
    request: Request,
    ns: Namespace,
    fmt: Rdf,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{ns.basename}.rdf"'
    return StreamingResponse(
        _stream_namespace(
            context,
            request,
            ns,
            action,
            data
        ),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


async def _stream(
    context: Context,
    request: Request,
    model: Model,
    action: Action,
    data
):
    namespaces = []
    prefixes = _get_available_prefixes(context, model)
    root_name = _get_attribute_name(RDF.upper(), RDF, prefixes)
    for key, val in prefixes.items():
        namespaces.append(f'xmlns:{key}="{val}"')
    if request.base_url:
        namespaces.append(f'xmlns="{str(request.base_url)}"')
    if isinstance(root_name, QName):
        root_name = f"{RDF}:{root_name.localname}"
    namespaces = "\n ".join(namespaces)

    yield f'<?xml version="1.0" encoding="UTF-8"?>\n<{root_name}\n {namespaces}>\n'
    if action == Action.GETONE:
        yield _prepare_for_print(model, prefixes, data)
    else:
        for row in data:
            yield _prepare_for_print(model, prefixes, row)
    yield f"</{root_name}>\n"


async def _stream_namespace(
    context: Context,
    request: Request,
    ns: Namespace,
    action: Action,
    data
):
    namespaces = []
    models = commands.traverse_ns_models(
        context,
        ns,
        ns.manifest,
        action,
        internal=True,
    )
    prefixes = {}
    used_datasets = []
    model_mapper = {}
    for model in models:
        prefixes = _get_update_prefixes(prefixes, used_datasets, model)
        model_mapper[model.name] = model
    prefixes = _update_required_prefixes(prefixes)

    root_name = _get_attribute_name(RDF.upper(), RDF, prefixes)
    for key, val in prefixes.items():
        namespaces.append(f'xmlns:{key}="{val}"')
    if request.base_url:
        namespaces.append(f'xmlns="{str(request.base_url)}"')
    if isinstance(root_name, QName):
        root_name = f"{RDF}:{root_name.localname}"
    namespaces = "\n ".join(namespaces)

    yield f'<?xml version="1.0" encoding="UTF-8"?>\n<{root_name}\n {namespaces}>\n'

    # TODO: Temporarily disabled page return, because page does not work with namespace/:all
    if action == Action.GETONE:
        model = model_mapper[data['_type'].text]
        if '_page' in data:
            data.pop('_page')
        yield _prepare_for_print(model, prefixes, data)
    else:
        for row in data:
            if '_page' in row:
                row.pop('_page')
            model = model_mapper[row['_type'].text]
            yield _prepare_for_print(model, prefixes, row)
    yield f"</{root_name}>\n"


@commands.prepare_data_for_response.register(Context, Model, Rdf, dict)
def prepare_data_for_response(
    context: Context,
    model: Model,
    fmt: Rdf,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    value = value.copy()
    reserved = get_model_reserved_props(action, page_in_data(value))

    available_prefixes = _get_available_prefixes(context, model)

    value['_available_prefixes'] = available_prefixes
    value['_about_name'] = _get_attribute_name('about', RDF, available_prefixes)
    value['_resource_name'] = _get_attribute_name('resource', RDF, available_prefixes)
    value['_type_name'] = _get_attribute_name('type', RDF, available_prefixes)
    value['_revision_name'] = _get_attribute_name('version', PAV, available_prefixes)

    data = {}
    for prop, val, sel in select_model_props(
        model,
        prop_names,
        value,
        select,
        reserved,
    ):
        prefix, name = _get_prefix_and_name(prop.uri)
        if name:
            value['_elem_name'] = _get_attribute_name(name, prefix, available_prefixes)
        else:
            value['_elem_name'] = _get_attribute_name(prop.name, prefix, available_prefixes)

        elem = commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel
        )
        data[prop.name] = elem
    return data


@commands.prepare_dtype_for_response.register(Context, Rdf, DataType, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: DataType,
    value: None,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return None


@commands.prepare_dtype_for_response.register(Context, Rdf, DataType, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, PageType, list)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: PageType,
    value: list,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=encode_page_values(value).decode('ascii')
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Date, datetime.date)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Date,
    value: datetime.date,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=value.isoformat()
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Time, datetime.time)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Time,
    value: datetime.time,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=value.isoformat()
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=value.isoformat()
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Number, Decimal)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Number,
    value: Decimal,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Ref, (dict, str, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Ref,
    value,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    super_ = commands.prepare_dtype_for_response[Context, Format, Ref, type(value)]
    data_dict = super_(context, fmt, dtype, value, data=data, action=action, select=select)
    prefixes = data['_available_prefixes']
    attributes = {}
    children = []
    if data_dict is None:
        return None

    if len(data_dict) == 1 and '_id' in data_dict or '_uri' in data_dict:
        if '_uri' in data_dict:
            attributes[data['_resource_name']] = data_dict['_uri'].text
        else:
            attributes[data['_resource_name']] = get_model_link(
                dtype.model,
                pk=data_dict['_id'].text
            )
    else:
        # TODO: write tests for denormalized form after task #216 is done

        if isinstance(data['_elem_name'], QName):
            prefix, name = _get_prefix_and_name(dtype.model.uri)
            if name:
                ref_model_name = _get_attribute_name(name, prefix, prefixes)
            else:
                ref_model_name = _get_attribute_name(dtype.model.basename, prefix, prefixes)
            data_dict, ref_model_attrs = _get_attributes(
                dtype.model,
                data_dict,
                data['_about_name'],
                data['_type_name'],
                data['_revision_name'],
            )
            ref_model_elem = _create_element(
                name=ref_model_name,
                attributes=ref_model_attrs,
                children=list(data_dict.values())
            )
            children.append(ref_model_elem)
        else:
            data_dict['_type'] = _create_element(
                name='_type',
                text=dtype.model.model_type()
            )
            data_dict, description_attrs = _get_attributes(
                dtype.model,
                data_dict,
                data['_about_name'],
                data['_type_name'],
                data['_revision_name'],
            )
            description_elem = _create_element(
                name=_get_attribute_name(DESCRIPTION, RDF, prefixes),
                attributes=description_attrs,
                children=list(data_dict.values())
            )
            children.append(description_elem)

    if attributes or children:
        return _create_element(
            name=data['_elem_name'],
            attributes=attributes,
            children=children
        )
    else:
        return None


@commands.prepare_dtype_for_response.register(Context, Rdf, File, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: File,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return None


@commands.prepare_dtype_for_response.register(Context, Rdf, File, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: File,
    value: Dict[str, Any],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value['_id'] is not None:
        attributes = {
            data['_about_name']: get_model_link(
                dtype.prop.model,
                pk=data['_id'],
                prop=dtype.prop.name
            )
        }
        return _create_element(
            name=data['_elem_name'],
            attributes=attributes
        )
    else:
        return None


@commands.prepare_dtype_for_response.register(Context, Rdf, Text, str)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Text,
    value: str,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Text, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Text,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    result = []
    for key, item in value.items():
        if item is not None:
            prefixes = data['_available_prefixes']
            attributes = None
            if key != "C":
                attributes = {
                    _get_attribute_name("lang", XML, prefixes): key
                }
            result.append(_create_element(
                name=data['_elem_name'],
                attributes=attributes,
                text=str(item)
            ))
    return result


@commands.prepare_dtype_for_response.register(Context, Rdf, ExternalRef, (dict, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: ExternalRef,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    super_ = commands.prepare_dtype_for_response[Context, Format, Ref, type(value)]
    data_dict = super_(context, fmt, dtype, value, data=data, action=action, select=select)
    prefixes = data['_available_prefixes']
    attributes = {}
    children = []
    if data_dict is None:
        return None

    if len(data_dict) == 1 and '_id' in data_dict or '_uri' in data_dict:
        if '_uri' in data_dict:
            attributes[data['_resource_name']] = data_dict['_uri'].text
        else:
            attributes[data['_resource_name']] = get_model_link(
                dtype.model,
                pk=data_dict['_id'].text
            )
    else:
        if isinstance(data['_elem_name'], QName):
            prefix, name = _get_prefix_and_name(dtype.model.uri)
            if name:
                ref_model_name = _get_attribute_name(name, prefix, prefixes)
            else:
                ref_model_name = _get_attribute_name(dtype.model.basename, prefix, prefixes)
            data_dict, ref_model_attrs = _get_attributes(
                dtype.model,
                data_dict,
                data['_about_name'],
                data['_type_name'],
                data['_revision_name'],
            )
            ref_model_elem = _create_element(
                name=ref_model_name,
                attributes=ref_model_attrs,
                children=list(data_dict.values())
            )
            children.append(ref_model_elem)
        else:
            data_dict['_type'] = _create_element(
                name='_type',
                text=dtype.model.model_type()
            )
            data_dict, description_attrs = _get_attributes(
                dtype.model,
                data_dict,
                data['_about_name'],
                data['_type_name'],
                data['_revision_name'],
            )
            description_elem = _create_element(
                name=_get_attribute_name(DESCRIPTION, RDF, prefixes),
                attributes=description_attrs,
                children=list(data_dict.values())
            )
            children.append(description_elem)

    if attributes or children:
        return _create_element(
            name=data['_elem_name'],
            attributes=attributes,
            children=children
        )
    else:
        return None


@commands.prepare_dtype_for_response.register(Context, Rdf, Geometry, BaseGeometry)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Geometry,
    value: BaseGeometry,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle Geometry dtype with BaseGeometry value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Geometry, WKBElement)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Geometry,
    value: WKBElement,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle Geometry dtype with WKBElement value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, BackRef, (dict, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: BackRef,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value is None:
        return None

    # TODO: implement appropriate prepare support to handle BackRef dtype with dict value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, BackRef, str)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: BackRef,
    value: str,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle BackRef dtype with str value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, ArrayBackRef, (tuple, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: ArrayBackRef,
    value: tuple,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value is None:
        return None

    # TODO: implement appropriate prepare support to handle ArrayBackRef dtype with tuple value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, ArrayBackRef, list)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: ArrayBackRef,
    value: list,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle ArrayBackRef dtype with list value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, JSON, (object, type(None), NotAvailable))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: JSON,
    value: object,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value is None or isinstance(value, NotAvailable):
        return None

    # TODO: implement appropriate prepare support to handle JSON dtype with object value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Object, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Object,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle Object dtype with dict value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Binary, bytes)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Binary,
    value: bytes,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=base64.b64encode(value).decode('ascii')
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Array, (list, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Array,
    value: List[Any],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value is None:
        return None

    # TODO: implement appropriate prepare support to handle Array dtype with list value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Array, tuple)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Array,
    value: tuple,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle Array dtype with tuple value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Inherit, (object, type(None), NotAvailable))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Inherit,
    value: object,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value is None or isinstance(value, NotAvailable):
        return None

    # TODO: implement appropriate prepare support to handle Inherit dtype with object value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Inherit, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Inherit,
    value: dict,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    # TODO: implement appropriate prepare support to handle Inherit dtype with dict value
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, UUID, (object, type(None), NotAvailable))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: UUID,
    value: object,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value is None or isinstance(value, NotAvailable):
        return None

    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Denorm, (object, type(None), NotAvailable))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Denorm,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None,
):
    return commands.prepare_dtype_for_response(
        context,
        fmt,
        dtype.rel_prop,
        value,
        data=data,
        action=action,
        select=select
    )



from __future__ import annotations

from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder
from spinta.datasets.backends.sql.backends.mariadb.components import MariaDB


class MariaDBQueryBuilder(SqlQueryBuilder):
    backend: MariaDB



from typing import List

import sqlalchemy as sa


def compare_push_state_rows(
    engine: sa.engine.Engine,
    table_name: str,
    rows: List[dict],
    order_by=None
):
    if order_by is None:
        order_by = ['id']

    with engine.connect() as conn:
        state_rows = conn.execute(
            sa.text(
                f'SELECT * FROM "{table_name}" ORDER BY {", ".join(order_by)}'
            )
        )

        for i, row in enumerate(state_rows):
            compare_row = rows[i]
            for key, value in compare_row.items():
                assert getattr(row, key) == value



from typing import Any, List, overload

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB, UUID

from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.constants import UNSUPPORTED_TYPES
from spinta.utils.sqlalchemy import Convention
from spinta.backends.postgresql.helpers import get_column_name
from spinta.backends.postgresql.helpers import get_pg_name
from spinta.backends.postgresql.helpers.changes import get_changes_table
from spinta.backends.postgresql.helpers.name import PG_NAMING_CONVENTION, get_pg_table_name, get_pg_column_name
from spinta.components import Context, Model
from spinta.manifests.components import Manifest
from spinta.types.datatype import DataType, PrimaryKey, Ref


@overload
@commands.prepare.register(Context, PostgreSQL, Manifest)
def prepare(context: Context, backend: PostgreSQL, manifest: Manifest, **kwargs):
    # Prepare backend for models.
    for model in commands.get_models(context, manifest).values():
        if model.backend and model.backend.name == backend.name:
            commands.prepare(context, backend, model, **kwargs)


@overload
@commands.prepare.register(Context, PostgreSQL, Model)
def prepare(context: Context, backend: PostgreSQL, model: Model, ignore_duplicate: bool = False, **kwargs):
    table_name = get_table_name(model)
    main_table_name = get_pg_name(table_name)
    if table_name in backend.tables and ignore_duplicate:
        return

    columns = []
    for prop in model.properties.values():
        # FIXME: _revision should has its own type and on database column type
        #        should bet received from get_primary_key_type() command.
        if prop.name.startswith('_') and prop.name not in ('_id', '_revision'):
            continue

        column: List[Any] | Any | None = commands.prepare(context, backend, prop, **kwargs)

        if isinstance(column, list):
            columns.extend(column)
        elif column is not None:
            columns.append(column)

    if model.unique:
        for constraint in model.unique:
            prop_list = []
            for prop in constraint:
                name = prop.name
                if isinstance(prop.dtype, Ref):
                    if commands.identifiable(prop):
                        name = f'{name}._id'
                    elif prop.dtype:
                        name = f'{name}.{prop.dtype.refprops[0].name}'
                prop_list.append(name)

            columns.append(sa.UniqueConstraint(*prop_list))

    # Create main table.
    pkey_type = commands.get_primary_key_type(context, backend)
    main_table = sa.Table(
        main_table_name, backend.schema,
        sa.Column('_txn', pkey_type, index=True),
        sa.Column('_created', sa.DateTime),
        sa.Column('_updated', sa.DateTime),
        *columns,
    )
    if main_table_name == 'country':
        pp(model.manifest.path)
    backend.add_table(main_table, model)
    # Create changes table.
    changelog_table = get_changes_table(context, backend, model)
    backend.add_table(changelog_table, model, TableType.CHANGELOG)


@overload
@commands.prepare.register(Context, PostgreSQL, DataType)
def prepare(context: Context, backend: PostgreSQL, dtype: DataType, **kwargs):
    if dtype.name in UNSUPPORTED_TYPES:
        return
    prop = dtype.prop
    name = get_column_name(prop)
    types = {
        'string': sa.Text,
        'date': sa.Date,
        'time': sa.Time,
        'datetime': sa.DateTime,
        'temporal': sa.DateTime,
        'integer': sa.Integer,
        'number': sa.Float,
        'boolean': sa.Boolean,
        'binary': sa.LargeBinary,
        'json': JSONB,
        'spatial': sa.Text,  # unsupported
        'image': sa.Text,  # unsupported
        'url': sa.String,
        'uri': sa.String,
        'denorm': sa.String,
        'uuid': UUID(as_uuid=True),
    }

    if dtype.name not in types:
        raise Exception(
            f"Unknown type {dtype.name!r} for property {prop.place!r}."
        )
    column_type = types[dtype.name]
    nullable = not dtype.required
    return sa.Column(name, column_type, unique=dtype.unique, nullable=nullable)


@commands.get_primary_key_type.register()
def get_primary_key_type(context: Context, backend: PostgreSQL):
    return UUID()


@overload
@commands.prepare.register(Context, PostgreSQL, PrimaryKey)
def prepare(context: Context, backend: PostgreSQL, dtype: PrimaryKey, **kwargs):
    pkey_type = commands.get_primary_key_type(context, backend)
    base = dtype.prop.model.base
    if base and commands.identifiable(base):
        return [
            sa.Column('_id', pkey_type, primary_key=True),
            sa.ForeignKeyConstraint(
                ['_id'], [f'{get_pg_table_name(get_table_name(base.parent))}._id'],
                name=PG_NAMING_CONVENTION[Convention.FK] % {
                    "table_name": get_pg_table_name(get_table_name(base.parent)),
                    "column_0_N_name": "_id"
                }
            )
        ]
    return sa.Column('_id', pkey_type, primary_key=True)



from __future__ import annotations

from spinta.datasets.backends.sql.backends.mssql.components import MSSQL
from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder


class MSSQLQueryBuilder(SqlQueryBuilder):
    backend: MSSQL



from __future__ import annotations

import dataclasses
from builtins import staticmethod
from typing import Any

from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import TypedDict
from typing import Union

from spinta.components import Component, Context
from spinta.components import Mode
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Store
from spinta.core.enums import Access
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.prefix.components import UriPrefix

if TYPE_CHECKING:
    from spinta.backends.components import Backend
    from spinta.datasets.components import Dataset
    from spinta.datasets.keymaps.components import KeyMap


class MetaDataContainer(TypedDict):
    ns: Dict[str, Namespace]
    dataset: Dict[str, Dataset]
    model: Dict[str, Model]


def get_manifest_object_names():
    return MetaDataContainer.__annotations__.keys()


class ManifestGiven:
    access: str = None


class Manifest(Component):
    type: str = None
    name: str = None
    keymap: KeyMap = None
    backend: Backend = None
    parent: Component = None
    store: Store = None
    _objects: MetaDataContainer = None
    path: str = None
    access: Access = Access.protected
    prefixes: Dict[str, UriPrefix]
    enums: Enums

    # Backends defined in the manifest.
    backends: Dict[str, Backend] = None

    # List of other source manifests used to populate nodes into the main
    # manifest.
    sync: List[Manifest]

    mode: Mode = Mode.internal

    given: ManifestGiven

    dynamic: bool = False

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return False

    def __init__(self):
        self.given = ManifestGiven()

    def __repr__(self):
        return (
            f'<{type(self).__module__}.{type(self).__name__}'
            f'(name={self.name!r})>'
        )

    def __eq__(self, other: Union[Manifest, str]):
        if isinstance(other, str):
            # A hack for tests, to make possible things like:
            #     assert manifest == """d | r | m | ..."""
            # This uses pytest_assertrepr_compare hook and compare_manifest to
            # eventually compare manifests in ascii table form.
            from spinta.testing.manifest import compare_manifest
            left, right = compare_manifest(self, other)
            return left == right
        else:
            super().__eq__(other)

    def get_objects(self) -> dict:
        return self._objects


NodeSchema = Optional[Dict[str, Any]]
ManifestSchema = Tuple[Any, NodeSchema]


@dataclasses.dataclass
class ManifestPath:
    type: str = 'tabular'
    name: str = None
    path: str = None
    file: IO = None
    prepare: str = None
